#!/usr/bin/python

import logging
import os
import optparse
import shutil
import sys
import subprocess

from common import *

from debian.changelog import Changelog

logging.basicConfig(level=logging.INFO)


def bzr_branch_package(loc, pkg):
    cmd = ['bzr', 'branch', loc, pkg]
    try:
        subprocess.check_output(cmd)
        logging.info('Branched %s from %s' % (pkg, loc))
    except subprocess.CalledProcessError, e:
        logging.error('Could not branch from: %s' % loc)
        logging.error(e)
        raise

    pkg_dir = os.path.join(os.getcwd(), package)

    orig_cwd = os.getcwd()
    os.chdir(pkg_dir)
    try:
        subprocess.check_output(['bzr', 'get-orig-source'])
    except subprocess.CalledProcessError, e:
        logging.error('Could not get-orig-source in: %s' % pkg_dir)
        logging.error(e)
        raise
    return pkg_dir


def dget_package(loc, pkg):
    orig_dirs = [f for f in os.listdir(os.getcwd()) if os.path.isdir(f)]

    cmd = ['dget', '-ux', loc]
    try:
        subprocess.check_output(cmd)
        logging.info('dget %s from %s OK.' % (pkg, loc))
    except subprocess.CalledProcessError, e:
        logging.error('Could not dget from: %s' % loc)
        logging.error(e)
        raise

    # try to figure out what directory the untar'd orig.tar left us
    new_dirs = [f for f in os.listdir(os.getcwd()) if os.path.isdir(f)]
    pkg_dir = [d for d in new_dirs if d not in orig_dirs][0]
    # cleanup everything other than the dir and orig.tar.gz
    for f in os.listdir(os.getcwd()):
        if not f.endswith('orig.tar.gz') and f != pkg_dir:
            os.unlink(f)

    return os.path.join(os.getcwd(), pkg_dir)


def current_version(changelog):
    f = open(changelog, 'r')
    changelog = Changelog(f)
    f.close()
    return str(changelog._blocks[0].version)


def version_back(changelog, back):
    f = open(changelog, 'r')
    changelog = Changelog(f)
    f.close()
    block = changelog._blocks[:back].pop()
    return str(block.version)


def ubuntu_release(changelog):
    f = open(changelog, 'r')
    changelog = Changelog(f)
    f.close()
    dist = changelog._blocks[0].distributions
    return str(dist).split('-')[0]


def bump_changelog(pkg, pkg_dir, release, version, message):
    cwd = os.getcwd()
    os.chdir(pkg_dir)
    cmd = ['dch', '--force-distribution', '-b', '-D', release,
           '--newversion', version, message]
    subprocess.check_output(cmd)
    os.chdir(cwd)


def build_source_package(pkg_dir, release, from_version):
    cwd = os.getcwd()
    os.chdir(pkg_dir)
    cmd = ['debuild', '-S', '-sa',
           '--changes-option=-DDistribution=%s' % release, '-v%s' % from_version]
    try:
        subprocess.check_output(cmd)
        rc = True
    except subprocess.CalledProcessError:
        logging.error('Could not build source package @ %s' % pkg_dir)
        rc = False
    os.chdir(cwd)
    return rc

logging.basicConfig(level=logging.INFO, format='%(message)s')

DEFAULT_MSG = 'New update for the Ubuntu Cloud Archive.'

usage = '''

%prog [options] lp:some/bzr/branch || http://some/source/package.dsc

Examples:

Backport nova using bzr from quantal-updates

./%prog -w ./nova -m "A cool new update for the UCA" \\
    -p nova lp:ubuntu/quantal-updates/nova

Backport a remote glance source package, preserving the last two changelog entries in .changes:

./%prog -w ./glance-m "A long overdue update." -v2 \\
    \t-p glance https://launchpad.net/ubuntu/+archive/primary/+files/glance_2012.2.3-0ubuntu2.dsc

'''
parser = optparse.OptionParser(usage=usage)
parser.add_option('-p', '--package', action='store', dest='package',
                  default=None, help='Package name.')
parser.add_option('-m', '--message', action='store', dest='message',
                  help='changelog message for new version (default: %s)' %\
                  DEFAULT_MSG, default=DEFAULT_MSG)
parser.add_option('-w', '--workdir', action='store', dest='workdir',
                  help='Working directory (default: cwd).',
                  default=None)
parser.add_option('-v', '--versions', action='store', dest='versions',
                  default=1, help='How many previous changelog versions '\
                  'to include in new .changes file. (default: 1)')
parser.add_option('-u', '--ubuntu_version', action='store', dest='ubuntu_vers',
                  default=None, help='Ubuntu Version backporting from (optional).')
parser.add_option('-c', '--cloud_version', action='store', dest='cloud_vers',
                  default=None, help='Cloud archive version tag (default: cloud0)')
(options, args) = parser.parse_args()

if len(args) == 0:
    logging.error('Must specify package dsc or bzr location to backport!')
    sys.exit(1)

if not options.package:
    logging.error('Must specify package location to backport!')
    sys.exit(1)

package = options.package
backport_loc = args.pop(0)


workdir = options.workdir or package

if os.path.isdir(workdir):
    user_confirm('Workdir (%s) exists, delete?' % workdir)
    shutil.rmtree(workdir)

os.mkdir(workdir)
os.chdir(workdir)
pref = backport_loc.split(':')[0]

pkg_dir = None

if pref in ['http', 'https', 'file']:
    build_type = 'dsc'
    pkg_dir = dget_package(backport_loc, package)
elif pref in ['lp', 'bzr', 'bzr+ssh']:
    build_type = 'bzr'
    pkg_dir = bzr_branch_package(backport_loc, package)
else:
    logging.error('Could not determine how to fetch from package location: '
                  '%s.' % backport_loc)
    sys.exit(1)

changelog = os.path.join(pkg_dir, 'debian', 'changelog')
cur_vers = current_version(changelog)
ca_vers = '%s~cloud0' % cur_vers

u_rel = options.ubuntu_vers or ubuntu_release(changelog)
os_rel = openstack_release(u_rel) or sys.exit(1)
ca_release = cloud_archive_release(os_rel) or sys.exit(1)
ca_pocket = '%s-%s' % (ca_release, os_rel)
ca_vers = '%s~cloud0' % cur_vers

version_from = version_back(changelog, int(options.versions) + 1)

bump_changelog(package, pkg_dir, ca_pocket, ca_vers, options.message)
build_source_package(pkg_dir, ca_release, version_from) or sys.exit(1)

shutil.rmtree(pkg_dir)
