from debian.changelog import Changelog
import re
import logging
import os

from launchpadlib.launchpad import Launchpad



# regular expressions needed to parse bugs from changelog
UBUNTU_BUG_RE = re.compile(r'\#(\d+)')
UPSTREAM_BUG_RE = re.compile(r'\LP: (\d+)')

# Ubuntu -> Openstack release
OS_UBUNTU_VERSION_MAP = {
    'precise': 'essex',
    'quantal': 'folsom',
    'raring': 'grizzly',
    'saucy': 'havana',
    'trusty': 'icehouse',
    'utopic': 'juno',
}

# Cloud Archive release -> Openstack version(s)
CA_OPENSTACK_VERSION_MAP = {
    'precise': ['folsom', 'grizzly', 'havana', 'icehouse'],
    'trusty': ['juno'],
}

# Supported LTS releases in the Cloud Archive.
SUPPORTED_CA = [
    'precise',
    'trusty',
]


def openstack_release(ubuntu_release):
    """
    Return the Openstack release shipped with a particular release of Ubuntu.

    :param ubuntu_release: Ubuntu release codename (eg, precise)
    """
    try:
        return OS_UBUNTU_VERSION_MAP[ubuntu_release]
    except KeyError:
        logging.error('Could not determine Openstack release for '\
                      'Ubuntu release %s' % ubuntu_release)
        return None


def cloud_archive_release(openstack_release):
    """
    Return the Ubuntu Release corresponding to the Cloud Archive pocket that
    provides a particular version of Openstack.

    :param openstack_release: Openstack release codename (eg, grizzly).
    """
    for k, v in CA_OPENSTACK_VERSION_MAP.iteritems():
        if openstack_release in v:
            return k
    logging.error('Could not determine CA release for OS release %s' %\
                  openstack_release)
    return None

def load_changelog(changelog_path):
    f = open(changelog_path, 'r')
    changelog = Changelog(f)
    f.close()
    return changelog

def changelog_top_block(changelog_path):
    '''returns the top block of a debian changelog'''
    cl = load_changelog(changelog_path)
    return cl._blocks[0]

def changelog_top_version(changelog_path):
    '''return the most recent version in changelog'''
    return changelog_top_block(changelog_path).version

def parse_changelog(path, version=None, ubuntu_bugs_only=True):
    """
    Parse bug #s from changelog at path.  By default, only the latest
    changelog entry will be parsed and only properly formatted Ubuntu
    bug references will be parsed, eg (LP: #XXXXXX).

    Returns a list of bug numbers.

    :param path: Path to changelog.

    :param version: Optional older package version.  If specified, bugs from
    all changelog entries since this version will be parsed.

    :param ubuntu_bugs_only: Only parse properly formatted Ubuntu bugs. 
    If False, upstream formatted bugs will be parsed as well, eg LP: #XXXXX.
    """
    changelog = open(path, 'r')
    changelog = Changelog(changelog)

    def _parse_changelog_entry(changelog_entry):
        entry = str(changelog_entry).splitlines()
        bugs = []

        sub = False
        sub_section = []
        anchor = '  * Resynchronize with'
        tabs = ['    - ', '      ']
        i = 0

        # Assemble sub-section of changelog.
        for line in entry:
            i += 1
            if line.startswith(anchor):
                sub = True
            if not sub:
                continue
            if re.search('\(LP: ', line):
                sub_section.append(line)
            if not ubuntu_bugs_only and re.search(' LP: ', line):
                sub_section.append(line)
            if True not in  [entry[i].startswith(t) for t in tabs]:
                sub = False

        for line in sub_section:
            [bugs.append(b) for b in UBUNTU_BUG_RE.findall(line)]
            if not ubuntu_bugs_only:
                [bugs.append(b) for b in UPSTREAM_BUG_RE.findall(line)]

        return bugs

    if version:
        logging.info('Parsing bugs since version %s' % version)
        exists = changelog_entry = None
        i = 0
        for b in changelog._blocks:
            i += 1
            if b.version == version:
                exists = True
                break
        if not exists:
            logging.error('Could not find %s in changelog' % version)
            raise Exception('No changelog entry.')
        entries = changelog._blocks[:i]

    else:
        entries =  [changelog._blocks[0]]
        logging.info('Parsing bugs from most recent version %s' %\
                         entries[0].version)
    bugs = []
    for e in entries:
        bugs += _parse_changelog_entry(e)

    return bugs


def extract_task(task):
    # taken from lp:ubuntu-qa-tools
    name = task.bug_target_name.lower()
    package = None
    distribution = None
    series = None
    if ' (' in name:
        package, distribution = name.split(' (',1)
        distribution = distribution.split(')')[0]
    else:
        package = name
    if distribution and ' ' in distribution:
        distribution, series = distribution.split(' ',1)
    return package, distribution, series


def ubuntu_task(bug):
    for task in bug.bug_tasks:
        e = extract_task(task)
        if e[1] == 'ubuntu':
            return True
    return False


def user_confirm(msg):
    while True:
        input = None
        input = raw_input('%s [Y|y]: ' % msg)
        if input in ['Y', 'y']:
            return True

LP = None
def get_lp():
    global LP
    if LP:
        return LP
    LP = Launchpad.login_anonymously('OpenStack Ubuntu Testing',
                                     'production')
    return LP
