#!/usr/bin/python

from launchpadlib.launchpad import Launchpad
from debian.changelog import Changelog
from lazr.restfulclient.errors import NotFound
import logging
import re
import optparse
import os
import sys

from common import *

logging.basicConfig(level=logging.INFO)

description="Parse Ubuntu bug numbers from the most recent block of a changelog. "\
            "For each bug in launchpad, open a task affecting the Ubuntu "\
            "package if it doesn't exist, mark Fixed Released (optional), "\
            "nominate for series (release), approve nomination and confirm "\
            "(assuming access)."
usage = "usage: %prog [options] /path/to/changelog"
parser = optparse.OptionParser(description=description, usage=usage)
parser.add_option('-r', '--release',
                  help='Ubuntu Release we are SRUing (default: precise)',
                  dest='target', action='store', default='precise')
parser.add_option('-c', '--component', help='OpenStack Component',
                  dest='component', action='store', default=None)
parser.add_option('-v', '--version', action='store', default=None,
                   dest='version', help='Act on all bugs since version.')
parser.add_option('-n', '--no-fix-release', action='store_true',
                  dest='no_fixed_rel', default=False,
                  help='By default, all new Ubuntu bug tasks or any bug task '\
                       'that isnt "Fix Released" will be marked as such '\
                       '(affecting Ubuntu dev. release)')
parser.add_option('-C', '--cloudarchive', action='store_true',
                  dest='cloud_archive', default=False,
                  help='Also set bugs affecting the ubuntu-cloud-archive '\
                       'project.')
(opts, args) = parser.parse_args()

if not opts.component:
    logging.error('Must specify Openstack Component')
    parser.print_help()
    sys.exit(1)

if opts.no_fixed_rel:
    mark_fixed_released = False
else:
    mark_fixed_released = True

if len(args) == 0:
    logging.error('Must specify path to changelog')
    parser.print_help()
    sys.exit(1)

changelog = args.pop(0)
if not os.path.exists(changelog):
    logging.error('Couldnt find changelog @ %s' % args[0])
    sys.exit(1)

target = opts.target
component = opts.component


lp = Launchpad.login_with('openstack-ubuntu-testing', 'production')
ubuntu = lp.distributions['ubuntu']
cur_dev = ubuntu.current_series.name

cur_tgt = ubuntu.getSeries(name_or_version=cur_dev)
sru_tgt = ubuntu.getSeries(name_or_version=target)

pkg = ubuntu.getSourcePackage(name=component)

def get_nomination(bug, target):
    try:
        return bug.getNominationFor(target=target)
    except NotFound as e:
        return False

def process_bug(bug_no):
    bug = lp.bugs[bug_no]
    if re.match('Meta bug', bug.title):
        logging.warn('Skipping bug %s: %s' % (bug_no, bug.title))
        return

    affects_ubuntu_package = False
    affects_component = False
    affects_cloud_archive = False

    def _set_fix_released(task):
        task.status = 'Fix Released'
        task.lp_save()

    for task in bug.bug_tasks:
        e = extract_task(task)

        # Special case a bunch of possible combos of project/distro/package
        if e[0] ==  'cloud-archive':
            affects_cloud_archive = True
        if e[0] == component:
            affects_component = True
        if e[0] == component and e[1] == 'ubuntu':
            affects_ubuntu_package = True
        if e[0] == 'ubuntu' and e[1:] == (None, None):
            # The bug task targetted at Ubuntu, but has no package.
            # Need to target pkg while we still have reference to task.
            logging.info('(LP: #%s) targets Ubuntu, but no pkg. Fixing.' % bug.id)
            task.target_link = pkg
            task.lp_save()
            e = (component, 'ubuntu', None)
        if e[0] == component and e[1] == 'ubuntu' and e[2] == None:
            if not task.status == 'Fix Released':
                logging.warn('(LP: #%s) is not marked "Fixed Released" in Ubuntu.' % bug.id)
                if mark_fixed_released:
                    logging.info('Setting existing task on (LP: #%s) "Fix Released"' % bug.id)
                    _set_fix_released(task)

    if not affects_component:
        logging.error('(LP: #%s) seems to not affect this component (%s)! Skipping.' % (bug.id, component))
        return False

    if not affects_cloud_archive and opts.cloud_archive:
        logging.info('Adding cloud-archive task on (LP: #%s).' % bug.id)
        ca_task = bug.addTask(target=lp.projects['cloud-archive'])
        ca_task.status = 'Confirmed'
        ca_task.lp_save()

    if not affects_ubuntu_package:
        logging.warn('(LP: #%s) does not currently target Ubuntu.' % bug.id)
        # this bug doesn't target ubuntu at all, open a task at the package,
        # optionally mark it fix released.
        logging.info('Created new task for (LP: #%s) affects %s' % (bug.id, target))
        new_task = bug.addTask(target=pkg)
        if mark_fixed_released:
            logging.info('Setting new bug task on (LP: #%s) to "Fix Released"' % bug.id)
            _set_fix_released(new_task)

    # Add nomination for the SRU target
    if bug.canBeNominatedFor(target=sru_tgt):
        logging.info('Nominating (LP: #%s) on %s for %s' % (bug_no, component, target))
        nomination = bug.addNomination(target = sru_tgt)
    else:
        nomination = None

    if not nomination:
        # bug may have already been nominated, grab its existing nomination.
        nomination = bug.getNominationFor(target=sru_tgt)

    if not nomination:
        # give up at this point.
        logging.error('Could not nominate (LP: #%s), giving up!' % bug_no)
        return False

    if nomination.status in ['Nominated', 'Declined']:
        logging.info('Approving nomination for (LP: #%s)' % bug_no)
        try:
            nomination.approve()
        except Exception as e:
            logging.error('Could not approve (LP: #%s), no permission?')

    # Confirm the new task
    for task in bug.bug_tasks:
        if extract_task(task) == (component, 'ubuntu', target):
            if task.status != 'Confirmed' and task.status != 'Fix Released':
                logging.info('Confirming new task on (LP: #%s)' % bug_no)
                task.status = 'Confirmed'
                task.lp_save()
            else:
                logging.info('Task already exists for (LP: #%s) and is Confirmed|Fix Released' % bug_no)
    return True

###
bugs = parse_changelog(changelog, version=opts.version)

logging.info('Parsed %s bugs from %s' % (len(bugs), changelog))

failed = []
success = []

for bug in bugs:
    try:
        if process_bug(bug):
            success.append(bug)
        else:
            failed.append(bug)
    except Exception as e:
        print e
        failed.append(bug)

print 'Successfully processed %s bugs' % len(success)

if failed:
    logging.error('Some bugs failed to be processed:')
    print '\n'.join([("(LP: #%s)" % i) for i in failed])

