#!/usr/bin/python

from debian.changelog import Changelog
from copy import copy
from launchpadlib.launchpad import Launchpad

import string
import json
import re
import os
import urllib
import sys
import logging
import tempfile
import tarfile
import optparse
import sys
import shutil

from common import *

parser = optparse.OptionParser()
parser.add_option('-p', '--package', dest='package', action='store',
                  default=None, help='Source package')
parser.add_option('-u', '--ubuntu-release', dest='ubuntu_release',
                  action='store', default='precise',
                  help='SRU Target Ubuntu Release (eg, precise)')
parser.add_option('-v', '--version', dest='version', action='store',
                  default=None,
                  help='Act on all bugs since version.')
parser.add_option('-f', '--file', dest='logfiles', action='append',
                   default=None, help='Test case log file.')
parser.add_option('-m', '--message', dest='message', action='store',
                  default=None, help='Verification message.')
parser.add_option('-F', '--force', dest='force', action='store_true',
                  default=False, help='Hands-free, force. No user interaction.')
parser.usage = 'usage: %s [options] /path/to/changelog\n\n'\
               'By default, script will post message, attach a test log and '\
               'flip verification tags.' % sys.argv[0]
(options, args) = parser.parse_args()

if not args:
    logging.error('Must pass path to changelog')
    exit(1)

changelog = args.pop(0)

if not os.path.exists(changelog):
    logging.error('Could not locate changelog @ %s' % changelog)
    exit(1)

if None in [options.package, options.logfiles, options.message]:
    logging.error('Missing arguments: package, logfile, message')
    exit(1)

for logfile in options.logfiles:
    if not os.path.isfile(logfile):
        logging.error('Could not find logfile @ %s.' % logfile)
        exit(1)

if not os.path.isfile(options.message):
    logging.error('Could not find verification message @ %s.' %\
                  options.message)
    exit(1)

package = options.package
version = options.version
openstack_release = OS_UBUNTU_VERSION_MAP[options.ubuntu_release]

# Owner link of upstream gerrit bot, who posts review results/commits
OPENSTACK_HUDSON_OWNER_URL = 'https://api.launchpad.net/1.0/~hudson-openstack'
# Commits to which stable branch do we care about.
OPENSTACK_STABLE_BRANCH = 'stable/%s' % openstack_release

logging.basicConfig(level=logging.INFO, format='%(message)s')

lp = Launchpad.login_with('openstack-ubuntu-testing', 'production')
ubuntu = lp.distributions['ubuntu']
archive = ubuntu.main_archive
series = ubuntu.getSeries(name_or_version=options.ubuntu_release)

def generate_msg(msg, context):
    text = open(msg, 'r').read()
    if context['master_review'] or context['stable_review']:
        text += "For records of upstream test coverage of this update, please "\
                "see the Jenkins links in the comments of the relevant "\
                " upstream code-review(s):\n\n"
        if context['master_review']:
            text +=  'Trunk review: %(master_review)s\n'
        if context['stable_review']:
            text += 'Stable review: %(stable_review)s\n'
        text += "\n"
    text +=  "As per the provisional Micro Release Exception granted to this package by "\
             "the Technical Board, we hope this contributes toward verification of this "\
             "update."

    return text % context


def get_bug_comments(bug_id):
    url = 'https://api.launchpad.net/1.0/bugs/%s/messages' % bug_id
    msgs = []
    bug = json.loads(urllib.urlopen(url).read())
    for m in bug['entries']:
        msg = {}
        msg['owner_link'] = m['owner_link']
        msg['date'] = m['date_created']
        msg['body'] = m['content']
        msgs.append(msg)
    return msgs

def display_comments(bug):
    for m in bug.comments:
        print m['body']

# These helpers are subject to change anytime upstream changes format
# of post-back msg.
def _has_review(body):
    body = body.splitlines()
    return body[0].startswith('Reviewed:')

def _get_committed_branch(body):
    body = body.splitlines()
    branch = body[3].split(': ')[1].strip()
    return branch

def _get_review_url(body):
    body = body.splitlines()
    url = body[0].split(': ')[1].split()
    return url

class LPBug(object):
    def __init__(self, bug):
        self.bug = bug
        self.id = self.bug.id
        self.title = self.bug.title
        self.comments = get_bug_comments(self.bug.id)
        self.stable_review = self._find_review(branch=OPENSTACK_STABLE_BRANCH)
        self.master_review = self._find_review()
        self.verification_msg = self._generate_verification_msg()
        self.tags = self.bug.tags

    def _find_review(self, branch='master'):
        url = None
        for c in self.comments:
            if c['owner_link'] == OPENSTACK_HUDSON_OWNER_URL and\
                _has_review(c['body']) and\
                _get_committed_branch(c['body']) == branch:
                url = _get_review_url(c['body'])
        if not url:
            return None
        return url[0]

    def _generate_verification_msg(self):
        context = {'project': package.title(),
                   'stable_review': self.stable_review,
                   'master_review': self.master_review}
        return generate_msg(options.message, context)

# MAIN

bugs = parse_changelog(changelog, version=version)

if not bugs:
    logging.error('Found no LP bug references in changelog entry since %s-%s.' %\
                    (package, version))

logging.info('Parsed from changelog and will act on the following bugs:')
lp_bugs = []

# Populate list from LP.
for bug_id in bugs:
        bug = lp.bugs[bug_id]
        if 'verification-done' not in bug.tags:
            lp_bug = LPBug(bug)
            lp_bugs.append(lp_bug)
        else:
            logging.info('Skipping (LP: #%s), already tagged '\
                         'verification-done' % bug_id)


def _clear():
    if not options.force:
        os.system('clear')

def _display(b,n=None):
    if n:
        num='%s. ' % n
    else:
        num=''
    logging.info('%s(LP: #%s) - %s' % (num, b.id, b.title))
    logging.info('\tstable_review: %s' % b.stable_review)
    logging.info('\tmaster_review: %s' % b.master_review)

def display_bugs(id=None):
    num = 1
    for bug in lp_bugs:
        _display(bug, num)
        num += 1

def edit_bug(bug):
    i = bug - 1
    bug = copy(lp_bugs[i])
    valid = ['M', 'm', 'S', 's', 'B', 'b', 'X', 'x']
    input = ''
    while True:
        _clear()
        while input not in valid:
            _display(bug)
            input = raw_input('[M]aster, [S]table, [B]ack, [X] Cancel: ')
        if input == 'M' or input == 'm':
            url = raw_input('Master review url: ')
            bug.master_review = url
            input = ''
        elif input == 'S' or input == 's':
            url = raw_input('Stable review url: ')
            bug.stable_review = url
            input = ''
        elif input == 'B' or input == 'b':
            lp_bugs.pop(i)
            lp_bugs.insert(i, bug)
            return
        elif input == 'X' or input == 'x':
            return

def post_messages(bugs=lp_bugs, logfiles=[]):
    for bug in bugs:
        bug.bug.newMessage(content=bug.verification_msg,
                           subject='Verification report.')
        logging.info('(LP: #%s): Msg added.' % bug.id)
        pkg_vers = changelog_top_version(changelog)

        for logfile in logfiles:
            _l = open(logfile, 'r')
            bug.bug.addAttachment(filename=os.path.basename(logfile),
                                  comment='Test coverage log.',
                                  data=open(logfile, 'r').read())
            logging.info('(LP: #%s): Attachment added.' % bug.id)

def revert_tags():
    for bug in lp_bugs:
        c = bug.tags.count('verification-done')
        if c:
            i = 1
            while i <= c:
                bug.tags.remove('verification-done')
                print "Tag removed"
                i += 1
            bug.tags.append('verification-needed')
            bug.bug.tags = bug.tags
            bug.bug.lp_save()

def flip_tags(bugs=lp_bugs):
    for bug in bugs:
        if 'verification-needed' in bug.tags:
            logging.info('(LP: %s): Removed verification-needed.' % bug.id)
            bug.tags.remove('verification-needed')
        if 'verificion-done' not in bug.tags:
            bug.tags.append('verification-done')
            logging.info('(LP: %s): verificaiton-done.' % bug.id)
        bug.bug.tags = bug.tags
        bug.bug.lp_save()

if options.force:
    display_bugs
    for bug in lp_bugs:
        post_messages([bug], options.logfiles)
        flip_tags([bug])
    sys.exit(0)

while True:
    _clear()
    logging.info('Parsed from changelog and will act on the following bugs:')
    display_bugs()
    valid=['E','e','R','r','M','m','T','t','d']
    input = ''
    while input not in valid:
        input = raw_input('([E]dit, [R]emove, Post [M]essages, Flip [T]ags)?: ')
    if input == 'E' or input == 'e':
        bug = raw_input('#: ')
        edit_bug(int(bug))
    elif input == 'R' or input == 'r':
        bug = raw_input('#: ')
        lp_bugs.pop(int(bug) - 1)
    elif input == 'M' or input == 'm':
        post_messages(logfiles=options.logfiles)
    elif input == 'T' or input == 't':
        flip_tags()
    elif input == 'd':
        import pdb; pdb.set_trace()
