#!/usr/bin/python
# Point this at a local git branch and generate a changelog containing
# all git commits between two periods, with bug refs.  Handy for preparing
# new SRUs from upstream OpenStack stable branches.
# TODO: Make use of git msg or LP bug title optional
#       Filter commits based on regexp
#       Commit & push?

import shutil
import argparse
import git
import re
import textwrap
import subprocess
import tempfile
import os
import logging
import sys
import time
import unicodedata
from copy import copy

import common

from launchpadlib.launchpad import Launchpad
launchpad = Launchpad.login_anonymously('just testing', 'production')

p = re.compile(r'\b([Bb]ug|[Ll][Pp])[\s:]*[#:]*\s*(\d+)') 

logging.basicConfig(level=logging.INFO)

def find_version(component, gittree):
    '''Attempt to obtain the current openstack version from the source.'''
    sys.path.insert(0, gittree)
    mod_name ='%s.version' % component
    try:
        vers_mod = __import__(mod_name, globals(), locals(), ['canonical_version_string'], -1)
        sys.path.pop(0)
    except ImportError:
        logging.warn('Could not import %s' % mod_name)
        return None
    vers = vers_mod.canonical_version_string()
    if vers:
        logging.info('Derived OpenStack version %s from source.' % vers)
        return vers
    return None

def parse_changes(entry):
    '''parse distro, version, changes out of parsechangelog output'''
    distro = version = changes = None
    i = 0
    entry = entry.split('\n')
    for l in entry:
        if l.startswith('Version'):
            version = l.split(':')[1].strip()
        elif l.startswith('Distribution'):
            distro = l.split(':')[1].strip()
        elif l.startswith('Changes:'):
            changes = '\n'.join(entry[i:])
        i += 1
    return distro, version, changes



def _process_commit(commit):
    skip_msgs = ['Bump version', 'Final versioning']

    # Defer to parents for all merge commits.
    if commit.summary.startswith("Merge ") and "into" in commit.summary:
        commit = commit.parents[1]

    msg = commit.message
    bugs = p.findall(msg)
    summary = commit.summary
    # skip 'release' commits
    for s in skip_msgs:
        if summary.startswith(s):
            return None
    # TODO: Use proper regexes
    # strip out bug numbers from description
    #summary =  re.sub(p, '', summary)

    summary = "[%s] %s" % (commit.hexsha[:7], summary )

    # if the commit has a bug number, lets append it
    if bugs:
        bugs = bugs[0][1]
        lpbug = launchpad.bugs[bugs]
        if common.ubuntu_task(lpbug):
            bugref = '(LP:$#%s)' % lpbug.id
        else:
            bugref = 'LP:$%s' % lpbug.id
        summary = "[%s] %s %s" % (str(commit)[:7], lpbug.title,  bugref)
        summary = unicodedata.normalize('NFKD', summary).encode('ascii','ignore')

    # format this so its wrapped and fits nicely as a
    # subsection of a changelog entry, ie:
    # * Resynchronize with stable/essex (272b98d7):
    #    - Entry summary...
    #      wrapped ending.
    summary = textwrap.wrap(summary,
                            80,
                            break_long_words=False,
                            initial_indent="    - ",
                            subsequent_indent="      ")
    out = []
    for line in summary:
        # Text wrap doesn't allow a word to contain a space, so
        # we hid it, now we re-add it.
        out.append(line.replace("LP:$", "LP: "))
    return '\n'.join(out)


def gen_changelog_entry(start, end=None, tree='.', branch='master'):
    repo = git.Repo(tree)

    for r in repo.heads:
        if str(r) == branch:
            target = r
            continue

    start_commit = repo.commit(start)

    if end:
        top_commit = repo.commit(end)
    else:
        top_commit = target.commit

    log = []
    log.append(_process_commit(top_commit))
    for commit in top_commit.iter_parents():
        summary = _process_commit(commit)
        if summary and summary not in log:
            log.append(summary)
        if commit == start_commit:
            break

    return log, top_commit

if __name__ == '__main__':
    orig_cwd = os.getcwd()
    parser = argparse.ArgumentParser(description='Create changelog for a given git commit range')
    parser.add_argument('--component', '-c', dest='component', help='Component')
    parser.add_argument('--os_release', '-O,',
                        dest='os_release', help='OpenStack Release. (default essex)',
                        default='essex')
    parser.add_argument('--os_version', '-v', help='OpenStack Version, '\
                        'ie 2012.1.2 (default auto)', dest='os_version',
                         default=None)
    parser.add_argument('--ubuntu_release', '-u', dest='distro_release',
                        help='Ubuntu Release (default precise)', default='precise')
    parser.add_argument('--start', '-s', dest='starthash', help='Hash of start')
    parser.add_argument('--end', '-e', dest='endhash', default=None,
                        help='Hash of end (default HEAD)')
    parser.add_argument('--gittree', '-g', dest='gittree', default='.',
                        help='location of git tree (default .)')
    parser.add_argument('--print_only', '-p', dest='print_only', 
                        action='store_true', default='False',
                        help="Don't bother updating changelog from bzr branch,"\
                             "only print git changelog and exit.")
    parser.add_argument('--branch', dest='gitbranch', default='stable/essex',
                        help='branch, such as "stable/essex" (default stable/essex)')
    parser.add_argument('--output', '-o', dest='output', default='changelog',
                        help='Output changelog. (default $PWD/changelog)')
    parser.add_argument('--meta_bug', '-m', dest='meta_bug', default=None,
                        help='Meta-bug # to add to entry header.')
    parser.add_argument('--timestamp', '-t', dest='timestamp', default=False,
                        action='store_true',
                        help='Add a timestamp to package version string. '\
                             '(default: False)')
    args = parser.parse_args()

    maintainer_email = os.getenv('DEBEMAIL')
    maintainer_name = os.getenv('DEBFULLNAME')

    if not maintainer_email or not maintainer_name:
        logging.error('Could not find DEBEMAIL or DEBFULLNAME in environment.')
        sys.exit(1)

    logging.info('Maintainer: %s <%s>' % (maintainer_name, maintainer_email))

    try:
        git_changes, end_commit  = gen_changelog_entry(start=args.starthash,
                                                       end=args.endhash,
                                                       tree=args.gittree,
                                                       branch=args.gitbranch)
    except Exception as e:
        logging.error('Could not generate git changes')
        logging.exception(e)
        sys.exit(1)

    # prune entires up to 'Bump stable/$foo next version'
    bump_found = False
    i = 0
    for ch in git_changes:
        if ch.split('] ')[1].startswith('Bump %s next version' %
                                        args.gitbranch):
            bump_found = i
        i += 1
    if bump_found:
        git_changes = git_changes[:bump_found]

    logging.info('Generated %s entries from git.' % len(git_changes))

    if args.print_only is True:
        print '\n'.join(git_changes)
        sys.exit(0)

    os_vers = args.os_version or find_version(args.component, args.gittree)
    if os_vers is None:
        logging.error('Could not determine base Openstack version from source.'\
                      'You can specify one explicity using -v')
        sys.exit(1)

    # Grab the changelog from the corresponding -proposed branch of
    # lp:openstack-ubuntu-testing
    tempdir = tempfile.mkdtemp()
    os.chdir(tempdir)
    if args.os_release == 'essex':
        lpb = 'lp:~openstack-ubuntu-testing/%s/%s-%s-proposed' %\
                 (args.component, args.distro_release, args.os_release)
    else:
        lpb = 'lp:~openstack-ubuntu-testing/%s/%s' %\
                 (args.component, args.os_release)

    cmd = ['bzr', 'branch', lpb, args.component]
    try:
        logging.info('Branching -proposed packaging in %s: %s' % (tempdir,
                                                                  ' '.join(cmd)))
        subprocess.check_call(cmd)
        os.chdir(args.component)
        changes = subprocess.check_output(['parsechangelog'])
        distro, version, changes = parse_changes(changes)
        changelog = open('debian/changelog', 'r').readlines()
    except Exception as e:
        logging.error('Could not branch from %s' % lpb)
        sys.exit(1)
    finally:
        shutil.rmtree(tempdir)

    short_hash = str(end_commit)[:8]

    if args.timestamp:
        date = time.strftime('%Y%m%d')
        new_vers = '%s+stable-%s-%s-0ubuntu1' % (os_vers, date, short_hash)
    else:
        new_vers = '%s-0ubuntu1' % os_vers

    entry_header = '%s (%s) UNRELEASED; urgency=low\n' % (args.component, new_vers)
    meta = ''
    if args.meta_bug:
        meta = ' (LP: #%s)' % args.meta_bug
    entry = '  * Resynchronize with %s (%s)%s:\n' % (args.gitbranch, short_hash, meta)
    for change in git_changes:
        entry += '%s\n' % change
    stamp = time.strftime('%a, %d %b %Y %H:%M:%S -0400')
    entry_footer = '\n -- %s <%s>  %s\n\n' % (maintainer_name, maintainer_email, stamp)

    # The -proposed branches should be kept as UNRELEASED until they're
    # uploaded.  If current version in pkg branch is UNRELEASED, its assumed
    # there are pending changes and its not a fresh changelog entry.
    new_changelog = os.path.join(orig_cwd, args.output) \
                    or  os.path.join(orig_cwd, 'changelog')
    if distro == 'UNRELEASED':
        logging.info('Current version in %s is UNRELEASED, appending to '\
                     'current entry instead of creating a fresh entry.' % (lpb))
        i = 0
        for l in changelog:
            if l.startswith(' -- '):
                footer = i
                i -= 1
                break
            i += 1

        new = changelog[:i]
        new[0] = entry_header
        new += '\n  [ %s ]\n' % maintainer_name
        new += entry
        new += entry_footer
        new += changelog[(i+3):]
        out = open(new_changelog, 'w+')
        out.writelines(new)
    else:
        new = open(new_changelog, 'w+')
        new.writelines([entry_header, '\n', entry, entry_footer] + changelog)
        new.close()

    logging.info('Wrote %s' % new_changelog)
