import shutil
import subprocess
import os
import yaml
import logging
import tempfile
import string
import re
import lockfile
import uuid

from openstack_ubuntu_testing.schroot.executor import SchrootExecutor

from time import strftime


class ComponentBuildException(Exception):
    ''' Marker exception for ComponentBuild issues '''
    pass


TARBALL_BASE_DEPS = [
    'python-setuptools',
    'git',
]

SOURCE_PACKAGE_BASE_DEPS = [
    'equivs',
    'devscripts',
    'bzr',
    'git-buildpackage',
    'git',
    'quilt',
    'gnupg',
]


NO_UCA = ('trusty-icehouse', 'xenial-mitaka')
DEFAULT_BRANCHES = ('upstream', 'pristine-tar')

DEFAULT_GIT_PKG_REPO = "https://git.launchpad.net/~ubuntu-server-dev/ubuntu/+source/{}"


class ComponentBuild():
    '''
    Class which supports building and installing Openstack packages
    for ubuntu based on upstream source locations
    '''
    def __init__(self, source, config,
                 project,
                 release=None,
                 dest=None,
                 branch=None):
        '''
        Constructor for ComponentBuild

        :param source: Source directory containing git clone of upstream
                       codebase.
        :param config: path to build configuration file (see etc).
        :param project: OpenStack project to build.
        :param release: Ubuntu release to build for.
        :param dest: Directory to store information in between builds.
        :param branch: Branch to create archive from - defaults to that
                       specified in the config file.
        '''
        self.source = os.path.abspath(source)
        with open(config, 'r') as f:
            self.config = yaml.load(f)
        self.project = project
        self.branch = branch
        self.schroot = None
        self.workdir = tempfile.mkdtemp()
        self.shell_log = []
        self.snapshot_name = "{}-{}".format(self.openstack_release,
                                            uuid.uuid4())
        self.release = release
        if dest:
            self.dest = os.path.abspath(dest)

        # NOTE: use defaults, then update with component specific bits
        self.component_config = self.config.get('component_defaults', {})
        self.component_config.update(
            self.config['components'][self.project] or {}
        )

    @property
    def openstack_release(self):
        return self.config['openstack_release']

    @property
    def target(self):
        return "{}-{}".format(self.release, self.openstack_release)

    def get_component_config(self, element):
        ''' Helper to interrogate yaml for component configuration '''
        return self.component_config.get(element)

    def get_release_config(self, element):
        ''' Helper to interrogate yaml for release configuration '''
        return self.config.get(element)

    def log_command(self, command):
        logging.debug(command)
        self.shell_log.append(string.join(command))

    def schroot_wrap(self, command, user=None):
        '''
        Executes the specified command in a schroot.  Performs schroot
        initialization if it hasn't been done already, ensuring any additional
        dependencies specified in release config get installed.
        '''
        # Setup schroot if not already created
        if not self.schroot:
            logging.info('Creating and configuring a new schroot for source'
                         'package/tar build')
            _base = self.get_release_config('additional_schroot_packages')
            self.schroot = SchrootExecutor(
                release='{}-amd64'.format(self.release),
                ubuntu_release=self.release,
                openstack_release=self.openstack_release,
                base_packages=_base
            )
            self.schroot.initialize_schroot()

        logging.info('Executing %s in schroot', command)
        self.log_command(command)
        self.schroot.execute(command, user)

    def install_tarball_build_dependencies(self):
            '''
            Install the minium dependencies required to build an sdist tarball.
            Some releases may define additional dependencies in their build
            release configs using the 'additional_schroot_packages' field.
            '''
            logging.info('Installing dependencies required for tarball build: '
                         + ' '.join(TARBALL_BASE_DEPS))
            cmd = ['apt-get', '-y', 'install'] + TARBALL_BASE_DEPS
            self.schroot_wrap(cmd, user='root')

    def install_source_build_dependencies(self):
            '''
            Install the base dependencies required to build a source package
            into the schroot.
            '''
            logging.info('Installing dependencies required for source package '
                         'build: %s' % ' '.join(SOURCE_PACKAGE_BASE_DEPS))
            cmd = ['apt-get', '-y', 'install'] + SOURCE_PACKAGE_BASE_DEPS
            self.schroot_wrap(cmd, user='root')

    def checkout_packaging(self, local_repo):
        '''Checkout packaging'''
        bzr_repo = self.get_component_config('bzr_repo')
        git_repo = (
            self.get_component_config('git_pkg_repo') or
            DEFAULT_GIT_PKG_REPO.format(self.project)
        )
        cmd = None
        if bzr_repo:
            logging.info("Branching %s to determine build-deps", bzr_repo)
            cmd = ['bzr', 'branch', bzr_repo, local_repo]
        elif git_repo:
            logging.info("Cloning %s to determine build-deps", git_repo)
            cmd = ['git', 'clone', git_repo]
            git_branch = self.get_component_config('git_pkg_branch')
            if git_branch:
                cmd += ['-b', git_branch]
            cmd += [local_repo]

        if cmd is not None:
            self.log_command(cmd)
            subprocess.check_call(cmd)
            if git_repo:
                for branch in DEFAULT_BRANCHES:
                    cmd = ['git', 'branch', '--track',
                           branch, 'origin/{}'.format(branch)]
                    self.log_command(cmd)
                    subprocess.check_call(cmd,
                                          cwd=local_repo)
        else:
            raise ValueError('Packaging bzr branch or git repository'
                             ' not provided')

    def search_for_tarball(self, directory):
        if not os.path.exists(directory):
            e = ('Could not find tarball version in non-existent '
                 'directory: %s' % directory)
            raise ComponentBuildException(e)
        for f in os.listdir(directory):
            archive_name = (self.get_component_config('upstream_name') or
                            self.project)
            if f.startswith(archive_name) and f.endswith('.tar.gz'):
                f_new = f.replace(archive_name, self.project)
                if f_new != f:
                    os.rename(os.path.join(directory, f),
                              os.path.join(directory, f_new))
                return f_new

        e = ('Could not find tarball for %s in directory: %s' %
             (self.project, directory))
        raise ComponentBuildException(e)

    def version_string_from_tarball(self, tarball_name):
        '''
        Extract the version string from upstream tarball filename
        '''
        version = re.sub('(%s-)|(.tar.gz)' % self.project, '', tarball_name)
        if not version:
            e = ('Could not derive upstream version from tarball: %s.' %
                 tarball_name)
            raise ComponentBuildException(e)
        logging.debug('Derived package version from tarball: %s' % version)
        return version

    def oslo_version(self):
        # Try component first for release overrides
        oslo_version = self.get_component_config('oslo_version')
        if not oslo_version:
            # Grab it from the release itself
            oslo_version = self.get_release_config('oslo_version')
        return str(oslo_version)

    def inject_oslo_environment_version(self):
        '''
        Possibly inject OSLO_VERSION into environment prior to sdist build
        '''
        version = self.get_component_config('version')
        if ('oslo' in version or 'OSLO_VERSION' in version):
            # Override sdist version with something sensible for
            # the distro
            os.environ['PBR_VERSION'] = self.oslo_version()

    def determine_build_versions(self, upstream_version):
        '''
        Given an upstream version (assumed to bederived from a tarball),
        determine the reletive version based on how this build's
        versioning is configured.

        We need to know two possibly identitical versions.  One for the
        orig.tar.gz and the other for the package itself., eg:

            nova_2011.1.orig.tar.gz   AND 2011.1
            nova_2013.1.2.orig.tar.gz AND 1:2013.1.2

        :param upstream_version:  The upstream version string.
        :returns : tuple: (tarball version, package version)
        '''
        version = self.get_component_config('version')
        tarball_vers = None
        pkg_vers = None
        if version in ['auto', 'gerrit']:
            # derive the upstream version directly from the tarball name
            tarball_vers = upstream_version
            pkg_vers = upstream_version
        elif 'AUTO_VERSION' in version:
            # use the derived version for the tarball version, and munge
            # package  version according to config.
            tarball_vers = upstream_version
            pkg_vers = version % {'AUTO_VERSION': upstream_version}
        elif 'oslo' in version or 'OSLO_VERSION' in version:
            # use the oslo version for the tarball version, and munge
            # package version according to config (fallback to oslo
            # version for both)
            oslo_version = self.oslo_version()
            tarball_vers = oslo_version
            if 'OSLO_VERSION' in version:
                pkg_vers = version % {'OSLO_VERSION': oslo_version}
            else:
                pkg_vers = oslo_version
        else:
            # directly from config
            tarball_vers = version
            pkg_vers = version

        if None in [tarball_vers, pkg_vers]:
            e = ('Could not determine build versions based on upstream '
                 'version. tarball_vers = %s, pkg_vers = %s.' %
                 (tarball_vers, pkg_vers))
            raise ComponentBuildException(e)

        return (tarball_vers, pkg_vers)

    def create_tarball(self):
        '''
        Create a upstream tarball for a component as defined in config.
        :returns: str: full path to tarball.
        '''
        git_branch = self.get_component_config('git_branch')
        date = strftime('%Y%m%d%H%M')

        if not self.branch:
            self.branch = git_branch

        logging.debug('Using git branch: %s' % self.branch)

        if not os.path.exists(self.source):
            logging.error('Source directory does not exist.')
            raise ComponentBuildException('Source directory does not exist.')

        # Make a copy of the source repo in the working dir
        # allows commands in schroot to see it.
        working_source = os.path.join(self.workdir, 'git', self.project)
        shutil.copytree(self.source, working_source)
        os.chdir(working_source)

        logging.info('Creating tarball using sdist')
        self.install_tarball_build_dependencies()
        self.inject_oslo_environment_version()

        # BUILD SDIST
        cmd = ['python', 'setup.py', 'sdist']
        logging.debug('About to call setup.py (%s) in schroot', cmd)
        self.schroot_wrap(cmd)

        # locate the tarball that setup.py built.
        dist_dir = os.path.join(os.getcwd(), 'dist')
        tarball = self.search_for_tarball(dist_dir)
        upstream_version = self.version_string_from_tarball(tarball)

        # determine the base version of the build, eg 1:2013.1.2
        _vers = self.determine_build_versions(upstream_version)
        tarball_vers, pkg_vers = _vers

        # generate a unique date stamp, eg git201306192001~precise
        stamp = '.{}'.format(date)
        if self.get_component_config('version') == 'gerrit':
            # Embed the gerrit change ID with the datestamp.
            change_id = os.getenv('GERRIT_CHANGE_ID')
            if not change_id:
                e = 'Missing GERRIT_CHANGE_ID in env.'
                raise ComponentBuildException(e)
            change_id = change_id[:8]
            stamp = '{}.gerrit.{}'.format(stamp, change_id)
            self.snapshot_name = '%s-gerrit-%s' % (self.project, change_id)
        stamp = '{}.{}'.format(stamp, self.release)

        # combine stamp with the base versions
        pkg_vers = "{}{}-0ubuntu1".format(pkg_vers, stamp)
        build_version = '{}{}'.format(tarball_vers, stamp)

        self.package_version = pkg_vers

        # move tarball out of src to the workspace
        package_name = self.project
        final_tarball = '%s_%s.orig.tar.gz' % (self.project,
                                               build_version)
        self.tarball = os.path.join(self.workdir, final_tarball)
        shutil.move('%s/dist/%s' % (working_source, tarball), self.tarball)


        self.source_dsc = ('%s_%s-0ubuntu1.dsc' %
                           (package_name, build_version))
        self.source_changes = ('%s_%s-0ubuntu1_source.changes' %
                               (package_name, build_version))
        self.amd64_changes = ('%s_%s-0ubuntu1_amd64.changes' %
                              (package_name, build_version))
        self.source_debtgz = ('%s_%s-0ubuntu1.debian.tar.gz' %
                              (package_name, build_version))
        return self.tarball

    def create_source_package(self, sign=True, key_id=None, bzr_branch=None,
                              cl_msg='auto'):
        '''
        Creates a source package using the configuration found in config.

        :param sign	     :   bool: Whether to GPG sign the source package
        :param keyid	 :    str: GPG key id to use if signing. If none, key ID
                                   will be pulled from build config.
        :param bzr_branch:    str: bzr packaging branch to build with.  If None
                                   repo will be pulled from build config.
        :param cl_msg    :    str: Changelog message.  If 'auto' it will be
                                   generated from git.
        :returns	     :  tuple: Full paths to source package components:
                                   (changes, dsc, tarball, amd64_changes,
                                    debtgz)
        '''
        gpg_key = key_id or self.get_release_config('gpg_key')

        os.chdir(self.workdir)

        if not os.path.isdir(self.source):
            logging.error("Source directory does not exist %s", self.source)
            raise ComponentBuildException("Source directory does not exist.")

        if os.path.isfile(self.tarball):
            logging.info("Building package using %s", self.tarball)
        else:
            logging.error("Unable to locate source tarball %s", self.tarball)
            raise ComponentBuildException("Unable to locate source tarball")

        # May have already been created by schroot_wrap
        if not os.path.isdir(os.path.join(self.workdir, self.project)):
            self.checkout_packaging(self.project)

        if os.path.exists(os.path.join(self.project, '.git')):
            cmd = ['gbp', 'import-orig',
                   '--no-interactive',
                   '--merge-mode=replace',
                   self.tarball]
            self.log_command(cmd)
            subprocess.check_call(cmd,
                                  cwd=os.path.join(self.workdir, self.project))

        os.chdir(self.project)
        logging.info("Generating new changelog entry")
        cmd = [
            'dch', '-b', '-D', self.release,
            '--newversion', self.package_version,
            'Automated Ubuntu testing build:'
        ]
        self.log_command(cmd)
        subprocess.check_call(cmd)

        if cl_msg == 'auto':
            os.chdir(self.source)
            logging.info("Generating git changelog entries for this package")
            entries = self.generate_changelog().split('\n')
            os.chdir(os.path.join(self.workdir, self.project))
        else:
            if isinstance(cl_msg, list):
                entries = cl_msg
            else:
                entries = [cl_msg]

        for entry in entries:
            cmd = ['dch', '-a', entry]
            self.log_command(cmd)
            subprocess.check_call(cmd)

        logging.info("Building source package")
        self.install_source_build_dependencies()
        cmd = ['debuild', '-S', '-sa', '-us', '-uc']
        self.log_command(cmd)
        self.schroot_wrap(cmd)

        if sign:
            os.chdir(self.workdir)
            logging.info("Signing source pacakge")
            cmd = ['debsign', '-k%s' % gpg_key, self.source_changes]
            self.log_command(cmd)
            subprocess.check_call(cmd)

        amd64_changes = os.path.join(self.workdir, self.source_dsc)
        changes = os.path.join(self.workdir, self.source_changes)
        dsc = os.path.join(self.workdir, self.source_dsc)
        dtgz = os.path.join(self.workdir, self.source_debtgz)
        return (changes, dsc, self.tarball, amd64_changes, dtgz)

    def build_package(self):
        ''' Builds the package using sbuild '''
        os.chdir(self.workdir)
        logging.info("Building package")
        if (self.package_version and
           os.path.isfile(self.source_dsc)):
            cmd = ['sbuild', '-d', self.release, '-n', '-A',
                   self.source_dsc]
            cmd += ['--chroot-setup-commands=apt-get -qy'
                    ' install software-properties-common']
            cmd += ['--chroot-setup-commands=add-apt-repository'
                    ' -y ppa:openstack-ubuntu-testing/{}'.format(self.openstack_release)]
            cmd += ['--chroot-setup-commands=apt-get update -q']
            cmd += ['--chroot-setup-commands=apt-get -y dist-upgrade']
            self.log_command(cmd)
            subprocess.check_call(cmd)
        else:
            e = "Unable to locate source package to build: %s", self.source_dsc
            logging.error(e)
            raise ComponentBuildException(e)

        return [f for f in os.listdir(self.workdir) if f.endswith('.deb')]

    def generate_changelog(self):
        ''' Generate changelog base on upstream commits '''
        os.chdir(self.source)
        cmd = [
            'git', 'log', '-n1', '--pretty=format:%H'
        ]
        self.log_command(cmd)
        self.current_commit = subprocess.check_output(cmd)
        logging.debug(self.current_commit)
        commit_store = os.path.join(self.dest,
                                    'previous.commit.%s' % self.project)
        if os.path.isfile(commit_store):
            logging.info("Detected previous commit - reading")
            with open(commit_store, 'r') as f:
                previous_commit = f.read().strip()
            if self.current_commit == previous_commit:
                return "No change rebuild."
            else:
                cmd = [
                    'git', 'log', '%s..HEAD' % previous_commit,
                    '--no-merges', '--pretty=format:[%h] %s'
                ]
                self.log_command(cmd)
                return subprocess.check_output(cmd)
        else:
            cmd = [
                'git', 'log', '-n5',
                '--no-merges', '--pretty=format:[%h] %s'
            ]
            self.log_command(cmd)
            return subprocess.check_output(cmd)

    def upload_source_package(self):
        ''' Uploads the build source package to the configured PPA '''
        ppa_url = self.get_release_config('ppa_url')

        if ppa_url == 'None' or ppa_url is None:
            return

        os.chdir(self.workdir)
        if os.path.isfile(self.source_changes):
            logging.info("Uploading package to %s", ppa_url)
            cmd = ['dput', ppa_url, self.source_changes]
            self.log_command(cmd)
            subprocess.check_call(cmd)
        else:
            e = "Unable to find source package to upload"
            logging.error(e)
            raise ComponentBuildException(e)

    def install_local_archive(self, archive=None, snapshot=False):
        ''' Installs the built packages using reprepro into archive '''
        os.chdir(self.workdir)
        if not archive:
            archive = self.get_release_config('local_archive')
        if not (os.path.isdir(archive) and
                os.path.isfile(self.amd64_changes)):
            msg = 'Unable to install file to local archive: Could not find '\
                  '%s or %s' % (archive, self.amd64_changes)
            logging.error(msg)
            raise ComponentBuildException(msg)

        repo_lock = lockfile.FileLock('/tmp/.repo_lock')
        with repo_lock:
            logging.info("Installing build artifacts into %s" % archive)
            cmd = [
                'reprepro', '--waitforlock', '10',  '-Vb', archive,
                'include', '%s-%s' % (self.release, self.openstack_release),
                self.amd64_changes
            ]
            self.log_command(cmd)
            subprocess.check_call(cmd)

            if snapshot:
                if not self.snapshot_name:
                    logging.info('Wanted a snapshot, but no snapshot name '
                                 'is set.  Is this a gerrit triggered job?')
                    return
                logging.info('Creating snapshot of repo with new package.')
                cmd = ['reprepro', '-b', archive, 'gensnapshot',
                       '%s-%s' % (self.release, self.openstack_release),
                       self.snapshot_name]
                self.log_command(cmd)
                subprocess.check_call(cmd)
                logging.info('Created new snapshot at %s-%s/snapshots/%s' %
                             (self.release, self.openstack_release,
                              self.snapshot_name))
                # Drop the snapshot path into a local file, which can then be
                # passed onto other jenkins jobs (deploy) so they know where
                # this file exists.  Wish there was a better way to handle
                # turning this into a build parameter. -adam_g
                snapshot_path = '%s-%s/snapshots/%s' %\
                                (self.release, self.openstack_release,
                                 self.snapshot_name)
                out = open('repo_snapshot', 'w')
                out.write('SNAPSHOT_REPO=%s' % snapshot_path)
                out.close()

    def store_commit(self):
        ''' Store the upstream commit for the current build on disk '''
        if self.current_commit:
            logging.info("Storing current commit for next build: %s" %
                         self.current_commit)
            commit_store = os.path.join(self.dest,
                                        'previous.commit.%s' % self.project)
            with open(commit_store, 'w') as f:
                f.write(self.current_commit)
        else:
            logging.info("No commit information to store")

    def cleanup(self):
        ''' Cleanup working directories etc.. '''
        shutil.rmtree(self.workdir)
        if self.schroot:
            logging.info('Destroying schroot.')
            self.schroot.cleanup_schroot()
            self.schroot = None

    def dump_shell_log(self):
        ''' Dump shell command log to stdout '''
        logging.info('Complete command log:')
        for cmd in self.shell_log:
            print('{}'.format(cmd))
