import ConfigParser
import logging
import yaml
import os
import sys
import uuid

# Current environment variables that serve as parameters for the
# test pipeline.
PIPELINE_PARAMS = [
    'UBUNTU_RELEASE',
    'OPENSTACK_RELEASE',
    'OPENSTACK_COMPONENT',
    'OPENSTACK_BRANCH',
    'PIPELINE_ID',
    'INSTALLATION_SOURCE',
    'DEPLOY_TOPOLOGY',
]

# Supported installation targets for cloud archive.
LTS_TARGETS = [
    'precise'
]

# Path to Juju deployer config containing available HW deployments.
DEPLOYMENT_CFG='%s/%s' %\
     (os.getenv('HOME'),
     'tools/openstack-ubuntu-testing/etc/deployer/deployments.cfg')

def load_pipeline_parameters(params='pipeline_parameters'):
    if not os.path.exists(params):
        logging.error('Could not find pipeline_parameters @ %s' % params)
        sys.exit(1)
    parameters = {}
    params = open(params, 'r').readlines()
    for l in params:
        if l.startswith('export'):
            l = l.split(' ')[1].strip()
            k, v = l.split('=')
            parameters[k.lower()] = v
    return parameters


#def determine_deployments(params):


def generate_manual_pipeline_params():
    params = {}

    ubuntu_rel = os.getenv('UBUNTU_RELEASE')
    os_rel = os.getenv('OPENSTACK_RELEASE')
    if None in [ubuntu_rel, os_rel]:
        logging.error('Did not specify UBUNTU_RELEASE or OPENSTACK_RELEASE')
        sys.exit(1)

    params['ubuntu_release'] = ubuntu_rel
    params['openstack_release'] = os_rel
    params['openstack_component'] = 'manual_trigger'
    params['openstack_branch'] = 'manual_trigger'
    params['pipeline_id'] = str(uuid.uuid4())
    params['installation_source'] = os.getenv('INSTALLATION_SOURCE') or\
                                    'default'
    params['deploy_topology'] = os.getenv('DEPLOY_TOPOLOGY') or 'default'
    return params

def generate_trunk_pipeline_params(job_name):
    if len(job_name.split('_')) != 4:
        logging.error('Imporperly formated job name: %s. Cannot derive '\
                      'pipeline parameters.' % job_name)
        sys.exit(1)
    params = {}
    params['ubuntu_release'] = job_name.split('_')[0]
    params['openstack_release'] = job_name.split('_')[1]
    params['openstack_component'] = job_name.split('_')[2]
    params['openstack_branch'] = job_name.split('_')[3]
    params['pipeline_id'] = str(uuid.uuid4())
    params['deploy_topology'] = 'default'
    params['installation_source'] = 'default'
    return params


def supported_deployment(params):
    vers_map = {
        'precise': 'essex',
        'quantal': 'folsom',
        'raring': 'grizzly',
        'saucy': 'havana',
        'trusty': 'icehouse',
        'utopic': 'juno',
        'vivid': 'kilo',
    }

    if 'cloud-archive' in params['installation_source']:
        if params['ubuntu_release'] not in LTS_TARGETS:
            logging.error('Cloud Archive installation only supported on: %s' %\
                           LTS_TARGETS)
            return False
        # cloud archive should support installation of everything.
        return True

    if params['ubuntu_release'] in vers_map and \
       vers_map[params['ubuntu_release']] == params['openstack_release']:
        return True
    else:
        # default (QA repo) and trunk-ppa provide what cloud archive would
        # typically satisfy.
        if params['installation_source'] in ['default', 'trunk-ppa']:
            return True
        else:
            logging.error('Unsupported OS + Ubuntu versions: %s on %s.' %
                          (params['openstack_release'],
                           params['ubuntu_release']))
            return False


def available_deployments():
    if not os.path.isfile(DEPLOYMENT_CFG):
        logging.error('Could not locate juju-deployer config: %s' %
                      DEPLOYMENT_CFG)
        sys.exit(1)
    return yaml.safe_load(open(DEPLOYMENT_CFG)).keys()


def deployments_for(base):
    '''
    return a list of available deployments that match a specific base name,
    eg, precise-folsom -> [precise-folsom, precise-folsom-quantum, ...]
    '''
    avail = available_deployments()
    deps = []
    for depl in avail:
        if depl.startswith(base):
            deps.append(depl)
    return deps


def openstack_origin(params):
    '''derive proper value for 'openstack-origin' juju charm config, based on
       pipeline params.'''
    src = params['installation_source']

    if src == 'default':
        return 'default'

    if src == 'trunk-ppa':
        return 'ppa:openstack-ubuntu-testing/%s-trunk-testing' %\
                params['openstack_release']

    if src in ['distro', 'distro-proposed']:
        return src

    if src.startswith('cloud-archive-'):
        ca_pocket = src.split('cloud-archive-')[1]
        if ca_pocket == 'updates':
            return 'cloud:%s-%s' % (params['ubuntu_release'],
                                    params['openstack_release'])
        elif ca_pocket == 'staging':
            return 'ppa:ubuntu-cloud-archive/%s-staging' %\
                    params['openstack_release']
        return 'cloud:%s-%s/%s' % (params['ubuntu_release'],
                                   params['openstack_release'],
                                   ca_pocket)


def enabled_services(enabled_services):
    '''return a list of all services enabled based on the generated
       ENABLED_SERVICES environment variable.'''
    services = []

    # mandatory
    services.append('keystone')
    services.append('nova')

    enabled_services = enabled_services.split(',')

    for service in enabled_services:
        if service in ['n-api', 'n-crt', 'n-obj', 'n-sch']:
            services.append('nova')
        if service in ['n-net']:
            services.append('nova-network')
        if service in ['g-api', 'g-reg', 'glance']:
            services.append('glance')
        if service in ['c-vol', 'c-api', 'c-sch', 'cinder']:
            services.append('cinder')
        if service in ['horizon']:
            services.append('horizon')
        if service in ['swift', 'swift-proxy', 'swift-storage']:
            services.append('swift')
        if service in ['quantum']:
            services.append('quantum')

    return list(set(services))

def determine_test_coverage(params, config_file=None):
    '''determine downstream test cases to schedule for given pipeline params
       and config_file.'''

    if not config_file:
        config_file = '%s/%s' % (os.getenv('HOME'),
                       'tools/openstack-ubuntu-testing/etc/tests/test_coverage.cfg')

    if not os.path.isfile(config_file):
        logging.error('Could not find test_coverage conf @ %s' % config_file)
        sys.exit(1)

    test_jobs = []

    config = ConfigParser.RawConfigParser()
    config.read(config_file)

    enabled = os.getenv('ENABLED_SERVICES')
    if enabled:
        services = enabled_services(enabled)
        # until tempest is up to snuff, only trigger the quantum test
        # for quantum deployments.
        if 'quantum' in services:
            test_jobs.append(config.get('PER_SERVICE_TESTS', 'quantum'))
            return test_jobs
        for service in services:
            if config.has_option('PER_SERVICE_TESTS', service):
                test_jobs.append(config.get('PER_SERVICE_TESTS', service))

    for job in config.get('DEFAULT', 'base_tests').split(','):
        test_jobs.append(job.strip())

    if config.has_option('PER_RELEASE_TESTS', params['openstack_release']):
        for job in config.get('PER_RELEASE_TESTS', params['openstack_release']).split(','):
            test_jobs.append(job.strip())

    return test_jobs
