'''
Library of useful functions for use in testing Openstack on Ubuntu
'''
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import os
import yaml
import logging
import jenkins
import sys
import copy

from jinja2 import Environment, FileSystemLoader


DEFAULT_GIT_REPO = "https://github.com/openstack/{}.git"


def create_precommit_pipeline(tmpl_dir, config_file, jenkins_url,
                              jenkins_username, jenkins_password, email,
                              release):

    '''
    Create a pre-commit testing pipeline.  This consists of a chain of
    dependent jobs that are different than the typical build jobs.

    For Precise Folsom Nova, this ends up looking something like:

    * gerrit trigger job that starts the pipeline and processes the results.
      - precise_folsom_precommit_nova_trigger
    * package build job that stuffs the builds into a reprepro snapshot
      specific to this test run.
      - precise_folsom_precommit_nova_pkg-build
    * juju deploy job that deploys the cloud using the snapshot (passed in
      as a build parameter)
      - precise_folsom_precommit_deploy
    * coverage job that preprares the cloud for testing, and triggers
      other jobs, one for each smoke test case.
      - precise_folsom_precommit_coverage
        -> precise_folsom_precommit_coverage_devstack
        -> precise_folsom_precommit_coverage_tempetst
        -> etc.

    The trigger job passes its build tag as a build parameter
    (PARENT_BUILD_TAG).
    This gets forwarded to all downstream jobs and its with this that we can
    later collate results for a specific run of the pipeline.

    Note that only the first two are specific to the component we're
    targetting. The later are generic and should be able to smoke test any
    build, so long as the correct build parameters are supplied.
    '''
    if not jenkins_url:
        logging.error('Please provide a URL to the jenkins instance')
        sys.exit(1)

    logging.debug('Using template directory %s' % tmpl_dir)

    templates = Environment(loader=FileSystemLoader(tmpl_dir))

    with open(config_file) as f:
        config = yaml.load(f)

    # Check to see if authentication information provided
    if jenkins_username:
        l_jenkins = jenkins.Jenkins(jenkins_url,
                                    username=jenkins_username,
                                    password=jenkins_password)
    else:
        l_jenkins = jenkins.Jenkins(jenkins_url)

    config = config[release]
    job_configs = {}

    ctxt = {
        'distro': release,
        'os_release': config['release'],
        'gerrit_host': config['gerrit_host'],
        'gerrit_user': config['gerrit_user'],
        'tests': config['test_cases']
    }

    for c in config['components']:
        comp = config.get('component_defaults', {})
        comp.update(config['components'][c])
        job_name = ('%s_%s_precommit_%s_trigger' %
                    (release, config['release'], c))
        ctxt['component'] = c
        ctxt['git_branch'] = comp['git_branch']

        # create the trigger job.
        job_name = ('%s_%s_precommit_%s_trigger' %
                    (release, config['release'], c))
        template = 'precommit_trigger.xml'
        logging.debug('Generating Jenkins Job %s' % job_name)
        template = templates.get_template(template)
        xml = template.render(ctxt)
        job_configs[job_name] = xml

        # create the pkg build job
        job_name = ('%s_%s_precommit_%s_pkg-build' %
                    (release, config['release'], c))
        template = 'precommit_pkg_build.xml'
        template = templates.get_template(template)
        xml = template.render(ctxt)
        job_configs[job_name] = xml

    # the deploy + coverage  jobs
    for j in ['deploy', 'coverage']:
        job_name = '%s_%s_precommit_%s' % (release, config['release'], j)
        template = templates.get_template('precommit_%s.xml' % j)
        xml = template.render(ctxt)
        job_configs[job_name] = xml

    # ensure the test cases all have jobs configured.
    # note, these jobs should be 100% generic,  run_tests.py should be able to
    # derive what tests to run just from the job name.
    template = templates.get_template('precommit_coverage_testcase.xml')
    xml = template.render()
    for t in config['test_cases']:
        job_name = ('%s_%s_precommit_coverage_%s' %
                    (release, config['release'], t))
        job_configs[job_name] = xml

    for l_jobname, j_config in job_configs.iteritems():
        if not l_jenkins.job_exists(l_jobname):
            logging.debug('Creating Jenkins Job %s ' % l_jobname)
            l_jenkins.create_job(l_jobname, j_config)
        else:
            logging.debug('Reconfiguring Jenkins Job %s ' % l_jobname)
            l_jenkins.reconfig_job(l_jobname, j_config)


def jenkins_login(jenkins_url, jenkins_username=None, jenkins_password=None):
    # Check to see if authentication information provided
    if jenkins_username:
        l_jenkins = jenkins.Jenkins(jenkins_url,
                                    username=jenkins_username,
                                    password=jenkins_password)
    else:
        l_jenkins = jenkins.Jenkins(jenkins_url)
    return l_jenkins


def create_test_pipeline(tmpl_dir, jenkins_url, jenkins_username,
                         jenkins_password, email):
    if not jenkins_url:
        logging.error('Please provide a URL to the jenkins instance')
        sys.exit(1)

    templates = os.path.join(tmpl_dir, 'test_pipeline')
    if not os.path.isdir(templates):
        logging.error('Could not find pipeline templates @ %s' % templates)

    templates = Environment(loader=FileSystemLoader(templates))

    l_jenkins = jenkins_login(jenkins_url, jenkins_username, jenkins_password)

    for job in ['runner', 'deploy', 'coverage', 'trigger']:
        tmpl = templates.get_template('%s.xml' % job)
        l_jobname = 'pipeline_%s' % job

        if not l_jenkins.job_exists(l_jobname):
            logging.debug('Creating Jenkins Job %s' % l_jobname)
            l_jenkins.create_job(l_jobname, tmpl.render())
        else:
            logging.debug('Reconfiguring Jenkins Job %s ' % l_jobname)
            l_jenkins.reconfig_job(l_jobname, tmpl.render())


def create_component_job(tmpl, tmpl_dir, config_file,
                         jenkins_url, jenkins_username, jenkins_password,
                         release=None, email=None):

    if not jenkins_url:
        logging.error('Please provide a URL to the jenkins instance')
        sys.exit(1)

    templates = Environment(loader=FileSystemLoader(tmpl_dir))

    with open(config_file) as f:
        config = yaml.load(f)

    l_jenkins = jenkins_login(jenkins_url, jenkins_username, jenkins_password)

    os_release = config['openstack_release']
    release = release or config['ubuntu_release']
    for component in config['components']:
        l_jobname = '%s_%s_%s' % (release, os_release,
                                  component)
        detail = copy.deepcopy(config.get('component_defaults', {}))
        detail.update(config['components'][component] or {})
        git_repo = detail.get('git_repo',
                              DEFAULT_GIT_REPO.format(component))
        git_branch = detail['git_branch']
        if 'gerrit_project' in detail:
            gerrit_project = detail['gerrit_project']
        else:
            gerrit_project = 'openstack/%s' % component
        ctxt = {'release': os_release,
                'distro': release,
                'branch': git_branch,
                'repo': git_repo,
                'gerrit_project': gerrit_project,
                'component': component,
                'email': email}
        logging.debug('Generating Jenkins Job %s ' % l_jobname)
        tmpl = templates.get_template(tmpl)
        j_config = tmpl.render(ctxt)
        logging.info('Creating/Reconfiguring Jenkins Job %s ' % l_jobname)
        if not l_jenkins.job_exists(l_jobname):
            logging.debug('Creating Jenkins Job %s ' % l_jobname)
            l_jenkins.create_job(l_jobname, j_config)
        else:
            logging.debug('Reconfiguring Jenkins Job %s ' % l_jobname)
            l_jenkins.reconfig_job(l_jobname, j_config)
