#!/usr/bin/python

import optparse
import logging
import os
import sys
import string
import subprocess
import re


def die(msg):
    logging.error(msg)
    sys.exit(1)


def list_configs(conf_dir):
    confs = []
    for f in os.listdir(conf_dir):
        if os.path.isdir(os.path.join(conf_dir, f)):
            confs += list_configs(os.path.join(conf_dir, f))
        if f.endswith('.conf'):
            confs.append(os.path.join(conf_dir, f))
    return confs


def update_test_configs(conf_dir):
    if not os.path.exists(conf_dir):
        die('Config directory does not exist at %s' % conf_dir)
    vars_file = os.path.join(conf_dir, 'all_variables')
    if not os.path.isfile(vars_file):
        die('Could not load all_variables from %s' % vars_file)

    env_vars = []
    for l in open(vars_file, 'r').readlines():
        if not l.startswith('#') and l.strip() != '':
            env_vars.append(l.strip())

    config_files = list_configs(conf_dir)

    for var in env_vars:
        if var in os.environ:
            val = os.environ[var]
            for f in config_files:
                sep = ","
                if re.search(",", val):
                    sep = "/"
                cmd = 'sed -i s%(sep)s%%%(var)s%%%(sep)s%(val)s%(sep)sg %(f)s' % locals()
                subprocess.check_output(cmd.split(' '))
        else:
            print "[WARN] Not setting %s in config, could not find in environment."\
                % var
    print "Configured tests in %s" % conf_dir

parser = optparse.OptionParser()
parser.add_option('-l', '--list', dest='list', action='store_true',
                  default=False, help='List available tests.')
parser.add_option('-c', '--config', dest='config', action='store',
                  default='run_tests.cfg', help='run_tests config file.')
parser.add_option('-j', '--jenkins', dest='jenkins', action='store_true',
                  default=False,
                  help='Attempt to determine test target from Jenkins'\
                       ' environment.')
parser.add_option('-C', '--configure', dest='configure', action='store_true',
                  default=False,
                  help='Configure all test configs based on enviornment.')
parser.add_option('-p', '--pdb', help='Drop into pdb on failure.',
                  dest='pdb_debug', action='store_true', default=False)
parser.add_option('-n', '--noschroot', help='Run tests outside of schroots',
                  dest='skip_schroot', action='store_true', default=False)
(options, args) = parser.parse_args()

if not os.path.isfile(options.config):
    die('Could not locate %s.' % options.config)


if options.configure:
    if len(args) == 0:
        die('Must pass a directory as argument to --configure')

    conf_dir=args.pop()
    update_test_configs(conf_dir)
    sys.exit(0)

import yaml

cfg = open(options.config, 'r')
config = yaml.load(cfg)

BASE_DEPS=None
if 'global_deps' in config:
    BASE_DEPS = config['global_deps']

AVAILABLE_TESTS = config['tests']

if options.list:
    print "Available tests:"
    for t in AVAILABLE_TESTS.keys():
        print "\t%s" % t
    sys.exit(0)

if options.jenkins:
    # Map Jenkins job names to test names.  Jenkins job names
    # should be formated as $distro_$release_coverage_$target.
    if 'JOB_NAME' not in os.environ:
        die('Could not find JOB_NAME in environment.')
    jenkins_job = os.environ['JOB_NAME'].split('_coverage_')
    if len(jenkins_job) != 2:
        die('JOB_NAME name not formatted properly, could not derive '\
            'test name.')
    test = string.join(jenkins_job, '_')
    print 'Test target dervied from Jenkins environment: %s' % test
else:
    if len(args) == 0:
        die('Must supply a test suite to run.')
    if len(args) > 1:
        die('Too many arguments passed.')
    test = args.pop()

if test not in AVAILABLE_TESTS.keys():
    die('Invalid test: %s' % test)

test_dir = AVAILABLE_TESTS[test]['dir']
if not os.path.exists(test_dir):
    die('Valid test specified, but cannot find test dir @ %s' % test_dir)

try:
    test_suite_type = AVAILABLE_TESTS[test]['suite_type']
except KeyError:
    test_suite_type = 'nose'

deps = AVAILABLE_TESTS[test]['base_deps']
# Append any additional global depends
if BASE_DEPS:
    for d in BASE_DEPS:
        if d not in deps:
            deps.append(d)

release = AVAILABLE_TESTS[test]['release']

if test_suite_type == 'nose':
    test_dir = AVAILABLE_TESTS[test]['dir']
    if not os.path.exists(test_dir):
        die('Valid test specified, but cannot find test dir @ %s' % test_dir)
    cmd = 'nosetests -sv --with-xunit %s' % test_dir
elif test_suite_type == 'shell':
    cmd = AVAILABLE_TESTS[test]['exec']
else:
    die('Invalid test suite type specified: %s' % test_suite_type)

cmd = cmd.split(' ')

if options.skip_schroot:
    runner = subprocess.check_call
else:
    from openstack_ubuntu_testing.schroot.executor import SchrootExecutor
    schroot = SchrootExecutor(release=release, base_packages=deps)
    schroot.initialize_schroot()
    print 'Initialized schroot for %s' % release
    runner = schroot.execute

exit_code = 0
try:
    runner(cmd)
except:
    if options.pdb_debug:
        import pdb
        pdb.set_trace()
    logging.error('Test run failed.')
    exit_code = 1
finally:
    if not options.skip_schroot:
        schroot.cleanup_schroot()

sys.exit(exit_code)
