
import unittest
import logging
import os
import shutil
import time
import urllib

from openstack_ubuntu_testing.test_utils.juju import client

from tests.auto_pkg_test.config import AutoPkgTestConfig


class AutoPkgTestRunner(unittest.TestCase):
    # NOTE: setUp + tearDown need conversion to classmethods if this
    # grows to more than a single test.
    def setUp(self):
        super(AutoPkgTestRunner, self).setUpClass()
        logging.debug('Ensuring juju environment bootstrapped')
        self.config = AutoPkgTestConfig()
        self.service = self.config.get('component')
        self.timeout = self.config.get('test_run_timeout')
        self.constraints = self.config.get('constraints')

        # ubuntu rel corresponds to juju environment
        self.env = self.config.get('ubuntu_release')
        self.result = None

        self.artifact_dir = self.config.get('artifact_dir')
        if os.path.exists(self.artifact_dir):
            shutil.rmtree(self.artifact_dir)
        os.mkdir(self.artifact_dir)

        client.call(['destroy-environment', '-e', self.env])
        client.ensure_bootstrap(environment=self.env,
                                constraints=self.constraints)

        repo = self.config.get('charm_repo')
        client.deploy(environment=self.env, repo_dir=repo,
                      charm='auto-pkg-test', service=self.service,
                      config=self.config.get('charm_config'), args=['-u'])

        client.call(['expose', '-e', self.env, self.service])
        # charm runs tests in the start hook. wait for installed.
        client.wait_for_state(self.service, state='installed',
                              environment=self.env)
        self.unit_address = client.get_machine_unit(self.service)

    def tearDown(self):
        self._save_artifacts()
        # Keep environment around after run, to diagnose any Openstack failures
        #client.call(['destroy-environment', '-e', self.env])

    def _get_state(self):
        url = 'http://%s/results/state' % self.unit_address
        return urllib.urlopen(url).read().strip()

    def _get_result(self):
        if self.result:
            return self.result
        url = 'http://%s/results/%s/result' % \
              (self.unit_address, self.service)
        self.result = urllib.urlopen(url).read().strip()
        return self.result

    def _save_artifacts(self):
        '''
        grabs results published by runner and saves in artifacts directory.
        '''
        # result and the log
        with open(os.path.join(self.artifact_dir, 'result'), 'w') as out:
            out.write('%s\n' % self._get_result())

        url = 'http://%s/results/test_output.log' % self.unit_address
        req = urllib.urlopen(url)
        if req.getcode() == 200:
            log_out = os.path.join(self.artifact_dir, 'test_output.log')
            with open(log_out, 'w') as out:
                out.write(req.read())

        # tarball
        url = 'http://%s/results/%s/%s.tar.gz' % \
              (self.unit_address, self.service, self.service)
        req = urllib.urlopen(url)
        if req.getcode() != 200:
            logging.info('No results tarball to archive.')
            return
        tar_out = os.path.join(self.artifact_dir,
                               '%s-adt-log.tar.gz' % self.service)
        with open(tar_out, 'w') as out:
            out.write(req.read())

    def test_dep8_auto_pkg_test(self):
        # "service" is installed and running the dep8 test.
        # poll http://$unit/results/$state until it is finished then
        # check its result.
        valid_states = ['running', 'finished']
        i = 0
        while True:
            _st = self._get_state()
            if _st.split(' ')[0] not in valid_states:
                logging.error('Runner reported invalid state: %s' % _st)
                raise Exception
            if _st == 'finished':
                result = self._get_result()
                self.assertEquals(result, 'SUCCESS (0)')
                return
            self.assertLessEqual(i, self.timeout, 'Test run timed out!')
            time.sleep(1)
            i += 1
