import nose
import signal
import socket
import subprocess
import os
import logging
import unittest
import time
import tempfile
import telnetlib
import urllib
import uuid

from openstack_ubuntu_testing.test_utils.juju import client
import openstack_ubuntu_testing.test_utils.cloud as cloud

from openstack_ubuntu_testing.test_utils.config import CloudConfig
from tests.juju.config import JujuTestConfig

from nose.tools import timed

SERVICE_1 = 'service1'
SERVICE_2 = 'service2'

cloud_config = CloudConfig()

class JujuInstallError(Exception):
    def __init__(self, msg='Juju install hook error'):
        logging.error(msg)



class TimedOut(Exception):
    def __init__(self, msg='Timed out'):
        logging.error(msg)


def time_out(secs):
    def _time_out(function):
        def _timeout(signal, frame):
            raise TimedOut('Function timed out after %s sec.')

        def wrapped(*args, **kwargs):
            signal.signal(signal.SIGALRM, _timeout)
            signal.alarm(secs)
            result = None
            timed_out = False
            try:
                result = function(*args, **kwargs)
            except TimedOut:
                timed_out = True
            finally:
                signal.alarm(0)
            if timed_out:
                raise TimedOut
            return result
        return wrapped
    return _time_out


@timed(600)
def _ensure_bootstrap():
    '''Bootstrap environment if it is not already.  To be run from
       any setUpClass() whos tests exercise Juju.
       TODO(adam_g): Try to isolate this into its own test case.'''
    if client.is_bootstrapped():
        logging.info('Environment Already bootstrapped.')
        return True

    cmd = ['juju', 'bootstrap']
    logging.info('Boostrapping new environment')
    try:
        subprocess.check_output(cmd)    
    except:
        logging.error('Calling "juju bootstrap" failed!')

    while not client.is_bootstrapped():
        time.sleep(5)
    return True


class JujuTests(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.juju_config = JujuTestConfig()
        self.juju_config.generate_environment_config()
        self.juju_config.generate_ssh_keypair()
        _ensure_bootstrap()
        self.bootstrap_node = client.status()['machines']['0']['dns-name']
        self.repo_dir = tempfile.mkdtemp()
        self.release = self.juju_config.get('release')
        os.mkdir(os.path.join(self.repo_dir, self.release))

        cmd = ['bzr', 'branch', self.juju_config.get('testing_charm'),
               os.path.join(self.repo_dir, self.release, 'testing_charm')]
        subprocess.check_output(cmd)

    @timed(30)
    def test_001_bootstrap(self):
        '''Ensure running agent and instance after boootstrap.'''
        machine = client.status()['machines']['0']
        instance_state = machine['instance-state']
        agent_state = machine['agent-state']
        self.assertEquals(instance_state, 'running')
        self.assertEquals(agent_state, 'running')


    @timed(600)
    def test_002_deploy(self):
        '''Ensure deployed services successful start.'''
        client.deploy(repo_dir=self.repo_dir, charm='testingcharm',
                     service=SERVICE_1)
        client.deploy(repo_dir=self.repo_dir, charm='testingcharm',
                     service=SERVICE_2)
        deployed = [SERVICE_1, SERVICE_2]
        failed = []
        started = []
        while True:
            for service, address, state in client.services():
                if state == 'started' and service not in started:
                    started.append(service)
                if state == 'install-error' and service not in failed:
                    failed.append(service)
                # Fail test if either charm failed to install.
                self.assertEqual(len(failed), 0)
                if sorted(deployed) == sorted(started):
                    # Test pass.
                    return
            # Keep trying until timeout/fail.
            time.sleep(5)


    @timed(60)
    def test_003_unexposed_port_unreachable(self):
        '''Confirm that an unexposed port is unreachable.'''
        # Port should be filtered by security group/firewall, consider
        # a 5 sec timeout an unreachable port.
        @time_out(5)
        def _try_connect(host, port):
            connection = telnetlib.Telnet(host, port)

        for service, address, state in client.services():
            client.call(['unexpose', service])
            time.sleep(5)
            connection_failed = False
            try:
                _try_connect(address, 80)
            except TimedOut:
                connection_failed = True
            else:
                pass
            self.assertTrue(connection_failed)


    @timed(60)
    def test_004_exposed_port_reachable(self):
        '''Confirm that an exposed port is reachable.'''
        # 15 sec timeout should be enough to let cloud process
        # the secgroup call.
        @time_out(15)
        def _try_connect(host, port):
            connection = telnetlib.Telnet(host, port)

        for service, address, state in client.services():
            client.call(['expose', service])
            connection_failed = False
            try:
                _try_connect(address, 80)
            except TimedOut:
                connection_failed = True
            else:
                pass
            self.assertFalse(connection_failed)


    @timed(60)
    def test_005_exposed_service_produces_a_service(self):
        '''Confirm an exposed service produces a service (web server).'''
        # the testing charm lays down a file in its webroot that contains
        # the string 'juju_service', make sure we can get it and ensure the
        # has actually started something.
        for s in [SERVICE_1, SERVICE_2]:
            # should be exposed by previous test, but just in case...
            client.call(['expose', s])
        time.sleep(5)
        for s in [SERVICE_1, SERVICE_2]:
            addr = client.get_machine_unit(s)
            url = 'http://%s/test_005' % addr
            resp = urllib.urlopen(url).read().strip()
            self.assertEqual('juju_service', resp)


    @timed(60)
    def test_006_config_set(self):
        '''Ensure environment reflects charm config settings.'''
        def _get_remote(addr):
            url = 'http://%s/test_006' % addr
            return urllib.urlopen(url).read().strip()

        def _ensure_setting(value):
            for s in [SERVICE_1, SERVICE_2]:
                # via 'juju get'
                conf = client.get_config(s)['settings']
                self.assertEqual(conf['test-option']['value'], value)
                # and also via config-changed remotely.
                addr = client.get_machine_unit(SERVICE_1)
                self.assertEqual(_get_remote(addr), value)

        # First check the default.
        _ensure_setting('Default')
        # Ensure new settings propagate.
        new_value = str(uuid.uuid1())
        for s in [SERVICE_1, SERVICE_2]:
            client.call(['set', s, 'test-option=%s' % new_value])
        _ensure_setting(new_value)


    def test_007_destroy_environment(self):
        access_key = cloud_config.get('ec2', 'access_key')
        secret_key = cloud_config.get('ec2', 'secret_key')
        endpoint = cloud_config.get('ec2', 'ec2_url')
        instance_ids = client.instance_ids()
        ec2 = cloud.EC2CloudClient(access_key, secret_key, endpoint)
        client.destroy_environment()
        timeout = 90
        logging.debug('Waiting for 90s juju env. to tear down.')
        while len(ec2.instances_running(instance_ids)) > 0:
            timeout -= 1
            if timeout < 0:
                running = ec2.instances_running(instance_ids)
                logging.error('Timed out waiting for env. to destroy. '\
                              'Instances still running: %s' % running)
                self.assertEqual(running, 0)
            time.sleep(1)
