import json
import nose
import os
import logging
import shutil
import time
import tempfile
import unittest
import ConfigParser

import pprint

import openstack_ubuntu_testing.test_utils.juju.client as juju
import openstack_ubuntu_testing.test_utils.cloud as cloud

from openstack_ubuntu_testing.test_utils.config import CloudConfig
config = CloudConfig()

INST_RUN_TIMEOUT = 120

DEFAULT_CONFIG='etc/tests/openstack_upgrade.conf'

class OpenStackUpgradeConfig(object):
    def __init__(self, conf=DEFAULT_CONFIG):
        if os.path.isfile(conf):
            self.config = ConfigParser.SafeConfigParser()
            self.config.read(conf)
        else:
            logging.error('No config file specified!')
            raise Exception

    def get(self, item):
        return self.config.get('DEFAULT', item)

def load_snapshots(work_dir):
    snaps = {}
    for snap in ['pre', 'post']:
        snapshot = os.path.join(work_dir, '%s_snapshot.json' % snap)
        snapshot = open(snapshot, 'r').read()
        snaps[snap] = json.loads(snapshot)

    return snaps['pre'], snaps['post']

class OSUpgradeTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.config = OpenStackUpgradeConfig()
        self.maxDiff = None
        self.work_dir = tempfile.mkdtemp()
        self.ec2 = cloud.EC2CloudClient(config.get('ec2', 'access_key'),
                                          config.get('ec2', 'secret_key'),
                                          config.get('ec2', 'ec2_url'))
        self.osapi = cloud.OSAPICloudClient(config.get('osapi', 'username'),
                                            config.get('osapi', 'password'),
                                            config.get('osapi', 'tenant_name'),
                                            config.get('osapi', 'auth_url'))


    @classmethod
    def tearDownClass(self):
        if os.path.isdir(self.work_dir):
            shutil.rmtree(self.work_dir)

    def test_001_create_cloud_resources(self):
        '''Create cloud resources on previous OpenStack version.'''
        for api in ['ec2', 'osapi']:
            client = getattr(self, api)
            print 'creating instance'
            client.create_key_pair_and_instance()
            print 'creating floating ip'
            client.create_and_associate_floating_ip()
            print 'creating volume'
            client.create_and_attach_volume()


    def test_002_snapshot_cloud_resources_pre_upgrade(self):
        '''Snapshot running cloud's resources pre-upgrade.'''
        for api in ['ec2', 'osapi']:
            client = getattr(self, api)
            snapshot = cloud.create_snapshot(client)
            if not os.path.isdir(os.path.join(self.work_dir, api)):
                 os.mkdir(os.path.join(self.work_dir, api))
            save = os.path.join(self.work_dir, api, 'pre_snapshot.json')
            with open(save, 'wb') as out:
                out.write(json.dumps(snapshot))


    def test_003_initialize_openstack_upgrade(self):
        '''Perform OpenStack upgrade via juju charms.'''
        services = config.enabled_services()
        upgrade_src = 'openstack-origin=%s' % self.config.get('upgrade_repo')
        for service in services:
            juju.call(['set', service, upgrade_src])
        # sleep, let upgrade complete and relations settle.
        time.sleep(120)


    def test_004_confirm_openstack_upgrade(self):
        '''Confirm software upgrade successful.'''
        failed = []
        status = juju.status()
        for unit in [status['services'][s]['units']
                     for s in status['services']]:
            for k, v in unit.iteritems():
                if v['agent-state'] != 'started':
                    failed.append((k, v['agent-state']))
        self.assertEquals(failed, [])

    def test_005_snapshot_cloud_resources_post_upgrade(self):
        '''Snapshot running cloud's resources post-upgrade.'''
        for api in ['ec2', 'osapi']:
            client = getattr(self, api)
            snapshot = cloud.create_snapshot(client)
            if not os.path.isdir(os.path.join(self.work_dir, api)):
                 os.mkdir(os.path.join(self.work_dir, api))
            save = os.path.join(self.work_dir, api, 'post_snapshot.json')
            with open(save, 'wb') as out:
                out.write(json.dumps(snapshot))


    def test_006_confirm_images_ec2(self):
        '''Confirm post-upgarde images via EC2 API.'''
        pre_snap, post_snap = load_snapshots(os.path.join(self.work_dir, 'ec2'))
        self.assertEquals(pre_snap['images'], post_snap['images'])


    def test_007_confirm_key_pairs_ec2(self):
        '''Confirm post-upgarde key_pairs via EC2 API.'''
        pre_snap, post_snap = load_snapshots(os.path.join(self.work_dir, 'ec2'))
        self.assertEquals(pre_snap['key_pairs'], post_snap['key_pairs'])


    def test_008_confirm_instances_ec2(self):
        '''Confirm post-upgarde instances via EC2 API.'''
        pre_snap, post_snap = load_snapshots(os.path.join(self.work_dir, 'ec2'))
        self.assertEquals(pre_snap['instances'], post_snap['instances'])


    def test_007_confirm_floating_ips_ec2(self):
        '''Confirm post-upgarde floating_ips via EC2 API.'''
        pre_snap, post_snap = load_snapshots(os.path.join(self.work_dir, 'ec2'))
        self.assertEquals(pre_snap['floating_ips'], post_snap['floating_ips'])


    def test_008_confirm_volumes_ec2(self):

        '''Confirm post-upgarde volumes via EC2 API.'''
        pre_snap, post_snap = load_snapshots(os.path.join(self.work_dir, 'ec2'))
        self.assertEquals(pre_snap['volumes'], post_snap['volumes'])



    def test_009_confirm_images_osapi(self):
        '''Confirm post-upgarde images via OSAPI API.'''
        pre_snap, post_snap = load_snapshots(os.path.join(self.work_dir,
                                                          'osapi'))
        self.assertEquals(pre_snap['images'], post_snap['images'])


    def test_010_confirm_key_pairs_osapi(self):
        '''Confirm post-upgarde key_pairs via OSAPI API.'''
        pre_snap, post_snap = load_snapshots(os.path.join(self.work_dir,
                                                          'osapi'))
        self.assertEquals(pre_snap['key_pairs'], post_snap['key_pairs'])


    def test_011_confirm_instances_osapi(self):
        '''Confirm post-upgarde instances via OSAPI API.'''
        pre_snap, post_snap = load_snapshots(os.path.join(self.work_dir,
                                                          'osapi'))
        self.assertEquals(pre_snap['instances'], post_snap['instances'])


    def test_012_confirm_floating_ips_osapi(self):
        '''Confirm post-upgarde floating_ips via OSAPI API.'''
        pre_snap, post_snap = load_snapshots(os.path.join(self.work_dir,
                                                          'osapi'))
        self.assertEquals(pre_snap['floating_ips'], post_snap['floating_ips'])


    def test_013_confirm_volumes_osapi(self):
        '''Confirm post-upgarde volumes via OSAPI API.'''
        # TODO wait for volume creation before checking state.
        return
        pre_snap, post_snap = load_snapshots(os.path.join(self.work_dir,
                                                          'osapi'))
        self.assertEquals(pre_snap['volumes'], post_snap['volumes'])

    def test_014_destroy_cloud_resources(self):
        '''Destroy upgrade test cloud resources.'''
        for api in ['ec2', 'osapi']:
            client = getattr(self, api)
            client.teardown()

