#!/usr/bin/env bash

# Modified devstack floating_ip test to stress basic
# Quantum functionality circa Folsom.

# **floating_ips.sh** - using the cloud can be fun

# we will use the ``nova`` cli tool provided by the ``python-novaclient``
# package to work out the instance connectivity

echo "*********************************************************************"
echo "Begin DevStack Exercise: $0"
echo "*********************************************************************"

# This script exits on an error so that errors don't compound and you see
# only the first error that occured.
set -o errexit

# Print the commands being run so that we can see the command that triggers
# an error.  It is also useful for following allowing as the install occurs.
set -o xtrace

# Cherry-picked from devstack functions.
function get_field() {
    while read data; do
        if [ "$1" -lt 0 ]; then
            field="(\$(NF$1))"
        else
            field="\$$(($1 + 1))"
        fi
        echo "$data" | awk -F'[ \t]*\\|[ \t]*' "{print $field}"
    done
}

function is_set() {
    local var=\$"$1"
    if eval "[ -z \"$var\" ]"; then
        return 1
    fi
    return 0
}

function die_if_not_set() {
    (
        local exitcode=$?
        set +o xtrace
        local evar=$1; shift
        if ! is_set $evar || [ $exitcode != 0 ]; then
            set +o xtrace
            echo $@
            exit -1
        fi
    )
}

# Settings
# ========
ASSOCIATE_TIMEOUT=30
ACTIVE_TIMEOUT=120
BOOT_TIMEOUT=120
TERMINATE_TIMEOUT=30

PRIVATE_NETWORK_NAME="$QUANTUM_FIXED_NET_NAME"

# Keep track of the current directory
#hEXERCISE_DIR=$(cd $(dirname "$0") && pwd)
#TOP_DIR=$(cd $EXERCISE_DIR/..; pwd)

# Import configuration
source envrc

# Instance type to create
DEFAULT_INSTANCE_TYPE=${DEFAULT_INSTANCE_TYPE:-m1.small}

# Boot this image, use first AMi image if unset
DEFAULT_IMAGE_NAME=${DEFAULT_IMAGE_NAME:-$IMAGE_NAME}

# Security group name
SECGROUP=${SECGROUP:-quantum_test_secgroup}

# Launching a server
# ==================

# List servers for tenant:
nova --no-cache list

# Images
# ------

# Nova has a **deprecated** way of listing images.
nova --no-cache image-list

# But we recommend using glance directly
glance image-list

# Grab the id of the image to launch
IMAGE=$(glance image-list | egrep " $DEFAULT_IMAGE_NAME " | tail -n1 | get_field 1)

# Security Groups
# ---------------

# List of secgroups:
nova --no-cache secgroup-list

# Create a secgroup
if ! nova --no-cache secgroup-list | grep -q $SECGROUP; then
    nova --no-cache secgroup-create $SECGROUP "$SECGROUP description"
    if ! timeout $ASSOCIATE_TIMEOUT sh -c "while ! nova --no-cache secgroup-list | grep -q $SECGROUP; do sleep 1; done"; then
        echo "Security group not created"
        exit 1
    fi
fi
# Determinine instance type
# -------------------------

# List of instance types:
nova --no-cache flavor-list

INSTANCE_TYPE=`nova --no-cache flavor-list | grep $DEFAULT_INSTANCE_TYPE | get_field 1`
if [[ -z "$INSTANCE_TYPE" ]]; then
    # grab the first flavor in the list to launch if default doesn't exist
   INSTANCE_TYPE=`nova --no-cache flavor-list | head -n 4 | tail -n 1 | get_field 1`
fi

NAME="ex-quantum"

VM_UUID=`nova --no-cache boot --flavor $INSTANCE_TYPE --image $IMAGE $NAME --security_groups=$SECGROUP --nic net-id=$QUANTUM_FIXED_NET_ID | grep ' id ' | get_field 2`
die_if_not_set VM_UUID "Failure launching $NAME"

# Testing
# =======

# First check if it spins up (becomes active and responds to ping on
# internal ip).  If you run this script from a nova node, you should
# bypass security groups and have direct access to the server.

# Waiting for boot
# ----------------

# check that the status is active within ACTIVE_TIMEOUT seconds
if ! timeout $ACTIVE_TIMEOUT sh -c "while ! nova --no-cache show $VM_UUID | grep status | grep -q ACTIVE; do sleep 1; done"; then
    echo "server didn't become active!"
    exit 1
fi

# get the IP of the server
IP=`nova --no-cache show $VM_UUID | grep "$QUANTUM_FIXED_NET_NAME" | get_field 2`
die_if_not_set IP "Failure retrieving IP address"

# Security Groups & Floating IPs
# ------------------------------

if ! nova --no-cache secgroup-list-rules $SECGROUP | grep -q icmp; then
    # allow icmp traffic (ping)
    nova --no-cache secgroup-add-rule $SECGROUP icmp -1 -1 0.0.0.0/0
    if ! timeout $ASSOCIATE_TIMEOUT sh -c "while ! nova --no-cache secgroup-list-rules $SECGROUP | grep -q icmp; do sleep 1; done"; then
        echo "Security group rule not created"
        exit 1
    fi
fi

# List rules for a secgroup
nova --no-cache secgroup-list-rules $SECGROUP

# Allocate a Quantum Floating IP for the VM
FLOATING_IP_ID=$(quantum floatingip-create $QUANTUM_EXT_NET_ID | grep " id " | get_field 2)
die_if_not_set FLOATING_IP_ID "Failure creating floating IP"
FLOATING_IP=$(quantum floatingip-show $FLOATING_IP_ID | grep floating_ip_address | get_field 2)
die_if_not_set FLOATING_IP "Failure creating floating IP"

# determine port of running server.
Q_PORT_ID=$(quantum port-list | grep "$IP" | get_field 1)
die_if_not_set Q_PORT_ID "Could not determine quantum port for VM with IP $IP"

# add floating ip to our server
quantum floatingip-associate "$FLOATING_IP_ID" "$Q_PORT_ID" ||
  die "Failure adding floating IP $FLOATING_IP ($FLOATING_IP_ID) to $NAME"

# test we can ping our floating ip within ASSOCIATE_TIMEOUT seconds
if ! timeout 60 sh -c "while ! ping -c1 -w1 $FLOATING_IP; do sleep 1 ; done"; then
  echo "Could not ping instance via floating IP $FLOATING_IP."
  exit 1
fi

# dis-allow icmp traffic (ping)
nova --no-cache secgroup-delete-rule $SECGROUP icmp -1 -1 0.0.0.0/0 || die "Failure deleting security group rule from $SECGROUP"

quantum floatingip-disassociate $FLOATING_IP_ID || die "Failure disassociating floating IP $FLOATING_IP"
quantum floatingip-delete $FLOATING_IP_ID || die "Failure deleteing IP $FLOATING_IP"

# Shutdown the server
nova --no-cache delete $VM_UUID || die "Failure deleting instance $NAME"

# Wait for termination
if ! timeout $TERMINATE_TIMEOUT sh -c "while nova --no-cache list | grep -q $VM_UUID; do sleep 1; done"; then
    echo "Server $NAME not deleted"
    exit 1
fi

# Delete a secgroup
nova --no-cache secgroup-delete $SECGROUP || die "Failure deleting security group $SECGROUP"

set +o xtrace
echo "*********************************************************************"
echo "SUCCESS: End DevStack Exercise: $0"
echo "*********************************************************************"
