import paramiko
import unittest
import os

from openstack_ubuntu_testing.test_utils.config import CloudConfig
cloud_config = CloudConfig()

logger = paramiko.util.logging.getLogger()
logger.setLevel('ERROR')

REMOTE_USER = 'ubuntu'

topology = {
        'nova-cloud-controller': {
            'host': cloud_config.get('DEFAULT', 'nova_host'),
            'processes': ['nova-api', 'nova-scheduler', 'nova-objectstore'],
            'ssh': None,
        },
        'nova-compute': {
            'host': cloud_config.get('DEFAULT', 'nova_compute_host'),
            'processes': ['nova-api', 'nova-network', 'nova-compute'],
            'ssh': None,
        },
        'nova-volume': {
            'host': cloud_config.get('DEFAULT', 'nova_volume_host'),
            'processes': ['nova-volume'],
            'ssh': None,
        },
        'glance': {
            'host': cloud_config.get('DEFAULT', 'glance_host'),
            'processes': ['glance-api', 'glance-registry'],
            'ssh': None,
        },
        'keystone': {
            'host': cloud_config.get('DEFAULT', 'keystone_host'),
            'processes': ['keystone'],
            'ssh': None,
        },
        'cinder': {
            'host': cloud_config.get('DEFAULT', 'cinder_host'),
            'processes': ['cinder-api', 'cinder-scheduler', 'cinder-volume'],
            'ssh': None,
        },
}


def _test_error(msg):
    print '[ERROR] %s' % msg


# This test case ensures each openstack process is running
# after deployment.  Its mostly meant to catch:
#   https://bugs.launchpad.net/bugs/1018586
class ServicesTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.connections = {}
        missing = []
        # Ensure we have all host info in our env, should have been
        # sourced in from envrc.
        for s in topology:
            if not topology[s]['host']:
                _test_error('Host missing from environment: %s' % s)
                missing.append(s)
        if missing:
            raise Exception('Hosts missing from enviornment.')
        # Open an ssh connection to each.
        failed = []
        for s in topology:
            topology[s]['ssh'] = paramiko.SSHClient()
            topology[s]['ssh'].set_missing_host_key_policy(paramiko.AutoAddPolicy())
            try:
                topology[s]['ssh'].connect(topology[s]['host'],
                                           username=REMOTE_USER)
            except:
                failed.append(s)
        if failed:
            raise Exception('Could not open ssh connection to hosts: %s' %\
                 failed)
        # Turn off paramiko logging at this point.
        paramiko.util.log_to_file('/dev/null')


    @classmethod
    def tearDownClass(self):
        for s in topology:
            topology[s]['ssh'].close()


    def _run_cmd(self, service, cmd):
        stdin, stdout, stderr = topology[service]['ssh'].exec_command(cmd)
        return (stdout.read(), stderr.read())


    def service_status(self, service):
        failed = []
        for process in topology[service]['processes']:
            res = self._run_cmd(service, 'status %s' % process)
            # Catch bad upstart jobs.
            if res[1].startswith('status: Unknown job'):
                _test_error('Upstart reports Unknown job: %s:%s' %\
                           (service, process))
                failed.append(process)
            # Ensure service is running.
            if not res[0].startswith('%s start/running' % process):
                _test_error('Upstart reports stopped service: %s:%s' %\
                           (service, process))
                # Job has failed, but see if we can catch any orphaned child
                # processes, see (LP: #1029727)
                proc = self._run_cmd(service, 'pgrep %s' % process)[0]
                if proc != '':
                    _test_error('Upstart reports job stopped for %s:%s, but '\
                               'found running processes: %s' %\
                                (service, process, proc))
                failed.append(process)
        return not failed


    def test_services_glance(self):
        '''Test running services on glance host'''
        self.assertTrue(self.service_status('glance'))


    def test_services_compute(self):
        '''Test running services on nova-compute host'''
        self.assertTrue(self.service_status('nova-compute'))


    def test_services_controller(self):
        '''Test running services on nova-cloud-controller host'''
        self.assertTrue(self.service_status('nova-cloud-controller'))


    def test_services_keystone(self):
        '''Test running services on keystone host'''
        self.assertTrue(self.service_status('keystone'))


    def test_services_cinder(self):
        '''Test running services on cinder host'''
        self.assertTrue(self.service_status('cinder'))

