#!/bin/bash -e
#
# Clone and run the Tempest test suite.
# This assumes all environment-specific data to use any clients has been
# sourced from envrc and is available in the environment.

function log() {
  echo "[$SCRIPT] $@"
}

function die() {
  echo "[ERROR] $@"
  cleanup
  exit 1
}

function setup() {
  TEST_DIR=$(dirname $0)
  SCRIPT=$(basename $0)
  WORKSPACE=$(pwd)
  TEMPEST_DIR='/tmp/tempest-test'

  if ! echo $TEMPEST_BRANCH | grep -q master ; then
    rel="$(echo $TEMPEST_BRANCH | cut -d/ -f2)"
    TEMPEST_CONF="tempest.$rel.conf"
  else
    TEMPEST_CONF="tempest.master.conf"
  fi
  TEMPEST_CONF="$WORKSPACE/etc/tests/tempest_templates/$TEMPEST_CONF"

  [[ -d $TEMPEST_DIR ]] && rm -rf $TEMPEST_DIR
  mkdir -p $TEMPEST_DIR && TEMPEST_DIR="$TEMPEST_DIR/tempest"

  [[ ! -e $TEMPEST_CONF ]] &&
    die "Could not locate generated Tempest config @ $TEMPEST_CONF"

  # Call any setup we may need using run-parts.
  if [[ -d $TEST_DIR/setup ]] ; then
    log "Running setup scripts in $TEST_DIR."
    run-parts --exit-on-error $TEST_DIR/setup
  fi
}

function cleanup() {
  # Call any setup we may need using run-parts.
  rm -rf $TEMPEST_DIR
  if [[ -d $TEST_DIR/cleanup ]] ; then
    log "Running cleanup scripts in $TEST_DIR."
    run-parts --exit-on-error $TEST_DIR/cleanup
  fi
}


CWD=$PWD

# envrc should be dropped in this directory prior to running.
. ./envrc
. ./etc/tests/tempest.conf

setup

TEMPEST_RUN_TESTS=${TEMPEST_RUN_TESTS:-$1}
# Default to running only the smoke tests, but can be overridde
# in either envrc or tempest.conf.  Currently, smoke tests are
# the only subset that is targetable.
case "$TEMPEST_RUN_TESTS" in
  ""|"smoke")
    attr_args="--attr=type=smoke"
    testr_args="smoke"
    ;;
  "scenario")
    tests_dir="$TEMPEST_DIR/tempest/scenario"
    testr_args="scenario";;
  "all")
    attr_args=""
    testr_args=""
    ;;
  "")
    attr_args="--attr=type=$TEMPEST_RUN_TESTS"
    testr_args="$TEMPEST_RUN_TESTS"
    ;;
esac

git clone $TEMPEST_REPO $TEMPEST_DIR
if [[ -n "$TEMPEST_BRANCH" ]] ; then
  (cd "$TEMPEST_DIR" && git checkout "$TEMPEST_BRANCH")
fi

cp $TEMPEST_CONF $TEMPEST_DIR/etc/tempest.conf

[[ -n "$tests_dir" ]] && TESTS_DIR="$tests_dir" ||
  TESTS_DIR="$TEMPEST_DIR/tempest"

if [ -f $TEMPEST_DIR/run_tempest.sh ]; then
    (cd $TEMPEST_DIR && testr init && testr run --parallel --subunit $testr_args | subunit2junitxml -o $CWD/nosetests.xml )
    rc=$?
else
    (cd $TEMPEST_DIR && nosetests --with-xunit --xunit-file=$CWD/nosetests.xml $attr_args -sv $TESTS_DIR)
    rc=$?
fi

cleanup

exit $rc
